from decouple import Csv, config as _config, undefined


def config(option: str, default=undefined, *args, **kwargs):
    if "split" in kwargs:
        kwargs.pop("split")
        kwargs["cast"] = Csv()

    if default is not undefined and default is not None:
        kwargs.setdefault("cast", type(default))
    return _config(option, default=default, *args, **kwargs)



# Generated by Django 2.0.13 on 2019-05-17 07:54

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CMISConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "client_url",
                    models.URLField(default="http://localhost:8082/alfresco/cmisatom"),
                ),
                ("client_user", models.CharField(default="admin", max_length=200)),
                ("client_password", models.CharField(default="admin", max_length=200)),
            ],
            options={"verbose_name": "CMIS Configuration"},
        )
    ]



# Generated by Django 2.2.10 on 2020-08-11 12:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("drc_cmis", "0009_auto_20200810_0830"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cmisconfig",
            name="base_folder_name",
            field=models.CharField(
                default="",
                help_text="Name of the DMS base folder in which the documents will be stored. "
                "If left empty, no base folder will be used.",
                max_length=200,
            ),
        ),
    ]



from django.conf import settings
from django.contrib import admin
from django.http import JsonResponse
from django.urls import path
from django.utils.translation import ugettext_lazy as _
from django.views import View

from solo.admin import SingletonModelAdmin

from .forms import CMISConfigAdminForm, UrlMappingInlineFormSet
from .models import CMISConfig, UrlMapping


class CMISConnectionJSONView(View):
    """Retrieve the vendor of the DMS behind the CMIS-connection."""

    model_admin = None

    def _get_status(self):
        """Retrieve the vendor from the CMIS repository."""
        cmis_config = CMISConfig.get_solo()
        if not cmis_config or not cmis_config.client_url:
            return _("N/A")

        try:
            from .client_builder import get_cmis_client

            return _("OK ({})").format(get_cmis_client().vendor)
        except KeyError:
            return _("Error: Unable to retrieve vendor)")
        except Exception as e:
            return _("Error: {}").format(e)

    def get(self, request, *args, **kwargs):
        """Return a JSON response showing the vendor or an error message."""
        if not self.has_perm(request):
            return JsonResponse({"error": "403 Forbidden"}, status=403)

        return JsonResponse({"status": self._get_status()})

    def has_perm(self, request, obj=None):
        """Check if user has permission to access the related model."""
        return self.model_admin.has_view_permission(request, obj=obj)


class UrlMappingAdmin(admin.TabularInline):
    model = UrlMapping
    formset = UrlMappingInlineFormSet
    can_delete = False
    extra = 1

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(CMISConfig)
class CMISConfigAdmin(SingletonModelAdmin):
    change_form_template = "configuration_form.html"
    form = CMISConfigAdminForm

    readonly_fields = [
        "cmis_connection",
        "cmis_url_mapping_enabled",
    ]
    fieldsets = [
        (
            _("General"),
            {"fields": ("cmis_connection",)},
        ),
        (
            _("DMS Credentials"),
            {
                "fields": (
                    "client_user",
                    "client_password",
                )
            },
        ),
        (
            _("Configuration"),
            {
                "fields": (
                    "client_url",
                    "binding",
                    "time_zone",
                    "main_repo_id",
                    "zaak_folder_path",
                    "other_folder_path",
                    "cmis_url_mapping_enabled",
                )
            },
        ),
    ]

    class Media:
        js = ("drc_cmis/js/cmis_config.js",)

    def get_urls(self):
        urls = super().get_urls()
        extra_urls = [
            path(
                "api/connection",
                self.admin_site.admin_view(self.cmis_connection_view),
                name="cmisconfig_api_connection",
            )
        ]
        return extra_urls + urls

    @property
    def inlines(self):
        if settings.CMIS_URL_MAPPING_ENABLED:
            return [UrlMappingAdmin]
        return []

    def cmis_connection_view(self, request):
        return CMISConnectionJSONView.as_view(model_admin=self)(request)

    def cmis_connection(self, obj=None):
        return "..."

    def cmis_url_mapping_enabled(self, obj=None):
        return settings.CMIS_URL_MAPPING_ENABLED

    cmis_connection.short_description = _("CMIS connection")

    cmis_url_mapping_enabled.short_description = _("URL Mapping enabled")
    cmis_url_mapping_enabled.boolean = True



import datetime
import logging
import uuid
from io import BytesIO
from typing import List, Optional, Union

from django.conf import settings

import pytz

from drc_cmis.models import CMISConfig
from drc_cmis.utils.exceptions import CmisRuntimeException, DocumentDoesNotExistError
from drc_cmis.utils.mapper import (
    DOCUMENT_MAP,
    GEBRUIKSRECHTEN_MAP,
    OBJECTINFORMATIEOBJECT_MAP,
    ZAAK_MAP,
    ZAAKTYPE_MAP,
    mapper,
)
from drc_cmis.utils.query import CMISQuery
from drc_cmis.utils.utils import extract_latest_version, get_random_string
from drc_cmis.webservice.data_models import (
    EnkelvoudigInformatieObject,
    Folder as _Folder,
    Gebruiksrechten as GebruiksrechtenDoc,
    Oio as OioDoc,
    QueriableUrl,
    ZaakFolderData,
    ZaakTypeFolderData,
    get_cmis_type,
    get_type,
)
from drc_cmis.webservice.request import SOAPCMISRequest
from drc_cmis.webservice.utils import (
    expand_url,
    extract_content,
    extract_object_properties_from_xml,
    extract_xml_from_soap,
    make_soap_envelope,
    pretty_xml,
    shrink_url,
)

logger = logging.getLogger(__name__)


class CMISBaseObject(SOAPCMISRequest):
    name_map = None
    type_name = None
    type_class = None

    def __init__(self, data):
        super().__init__()
        self.data = data
        self.properties = dict(data.get("properties", {}))

    def __getattr__(self, name: str):
        try:
            return super(SOAPCMISRequest, self).__getattribute__(name)
        except AttributeError:
            pass

        def resolve_attribute(name: str) -> str:
            if name in self.properties:
                return self.properties[name]["value"]

            convert_name = f"cmis:{name}"
            if convert_name in self.properties:
                return self.properties[convert_name]["value"]

            convert_name = f"drc:{name}"
            if self.name_map is not None and name in self.name_map:
                convert_name = self.name_map.get(name)

            if convert_name not in self.properties:
                raise AttributeError(f"No property '{convert_name}'")

            return self.properties[convert_name]["value"]

        value = resolve_attribute(name)

        if (
            get_type(self.type_class, name) == QueriableUrl
            and settings.CMIS_URL_MAPPING_ENABLED
            and value is not None
        ):
            return expand_url(value)

        return value

    @classmethod
    def build_properties(cls, data: dict) -> dict:
        """Construct property dictionary.

        The structure of the dictionary is (where ``property_name``, ``property_value``
        and ``property_type`` are the name, value and type of the property):

            .. code-block:: python

                properties = {
                    "property_name": {
                        "value": property_value,
                        "type": property_type,
                    }
                }
        """
        config = CMISConfig.objects.get()

        props = {}
        for key, value in data.items():
            prop_name = mapper(key, type=cls.type_name)
            if not prop_name:
                logger.debug("CMIS_ADAPTER: No property name found for key '%s'", key)
                continue
            if value is not None:
                prop_type = get_cmis_type(cls.type_class, key)

                if (
                    get_type(cls.type_class, key) == QueriableUrl
                    and settings.CMIS_URL_MAPPING_ENABLED
                ):
                    value = shrink_url(value)
                elif isinstance(value, datetime.datetime):
                    value = value.astimezone(pytz.timezone(config.time_zone)).strftime(
                        "%Y-%m-%dT%H:%M:%S.000Z"
                    )
                elif isinstance(value, datetime.date):
                    # In CMIS, there is no propertyDate, only propertyDateTime.
                    # So dates need to be in the datetime format
                    value = value.strftime("%Y-%m-%dT00:00:00.000Z")
                props[prop_name] = {"value": str(value), "type": prop_type}

        return props


class CMISContentObject(CMISBaseObject):
    def delete_object(self):
        """Delete all versions of an object"""

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=self.objectId,
            cmis_action="deleteObject",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

    def get_parent_folders(self) -> List["Folder"]:
        """Get all the parent folders of an object"""

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=self.objectId,
            cmis_action="getObjectParents",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "NavigationService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(
            xml_response, "getObjectParents"
        )
        return [Folder(data) for data in extracted_data]

    def move_object(self, target_folder: "Folder") -> "CMISContentObject":
        """Move a document to the specified folder"""

        source_folder = self.get_parent_folders()[0]

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=self.objectId,
            target_folder_id=target_folder.objectId,
            source_folder_id=source_folder.objectId,
            cmis_action="moveObject",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))
        extracted_data = extract_object_properties_from_xml(xml_response, "moveObject")[
            0
        ]

        return type(self)(extracted_data)

    def _update_properties(self, properties: dict) -> dict:
        """
        Update properties and return the properties of the updated object.

        :param properties: dict, new properties to update
        :return: dict, properties of the updated object
        """
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            properties=properties,
            cmis_action="updateProperties",
            object_id=self.objectId,
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))
        extracted_data = extract_object_properties_from_xml(
            xml_response, "updateProperties"
        )[0]

        return extracted_data

    def get_content_object(
        self, object_id: str, object_type: type
    ) -> "CMISContentObject":
        """Get a content object with specified objectId

        :param object_id: string, objectId of the content object
        :param object_type: type, type of the object to return
        :return: CMISContentObject
        """
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=object_id,
            cmis_action="getObject",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))
        extracted_data = extract_object_properties_from_xml(xml_response, "getObject")[
            0
        ]

        return object_type(extracted_data)


class Document(CMISContentObject):
    table = "drc:document"
    name_map = DOCUMENT_MAP
    type_name = "document"
    type_class = EnkelvoudigInformatieObject

    @classmethod
    def build_properties(
        cls, data: dict, new: bool = True, identification: str = ""
    ) -> dict:
        """Construct property dictionary.

        The structure of the dictionary is (where ``property_name``, ``property_value``
        and ``property_type`` are the name, value and type of the property):

            .. code-block:: python

                properties = {
                    "property_name": {
                        "value": property_value,
                        "type": property_type,
                    }
                }
        """

        config = CMISConfig.objects.get()

        props = {}
        for key, value in data.items():
            prop_name = mapper(key, type="document")
            if not prop_name:
                logger.debug("CMIS_ADAPTER: No property name found for key '%s'", key)
                continue
            if value is not None:
                prop_type = get_cmis_type(EnkelvoudigInformatieObject, key)

                if (
                    settings.CMIS_URL_MAPPING_ENABLED
                    and get_type(EnkelvoudigInformatieObject, key) == QueriableUrl
                    and value != ""
                ):
                    value = shrink_url(value)
                elif isinstance(value, datetime.datetime):
                    value = value.astimezone(pytz.timezone(config.time_zone)).strftime(
                        "%Y-%m-%dT%H:%M:%S.000Z"
                    )
                elif isinstance(value, datetime.date):
                    # In CMIS, there is no propertyDate, only propertyDateTime.
                    # So dates need to be in the datetime format
                    value = value.strftime("%Y-%m-%dT00:00:00.000Z")
                elif isinstance(value, bool):
                    value = str(value).lower()

                props[prop_name] = {"value": str(value), "type": prop_type}
            # When a Gebruiksrechten object is deleted, the field in the Document needs to be None.
            elif key == "indicatie_gebruiksrecht":
                prop_type = get_cmis_type(EnkelvoudigInformatieObject, key)
                props[prop_name] = {"value": "", "type": prop_type}

        # For documents that are not new, the uuid shouldn't be written
        props.pop(mapper("uuid"), None)

        if new:
            # increase likelihood of uniqueness of title by appending a random string
            title, suffix = data.get("titel"), get_random_string()
            if title is not None:
                props["cmis:name"] = {
                    "value": f"{title}-{suffix}",
                    "type": get_cmis_type(EnkelvoudigInformatieObject, "name"),
                }

            # make sure the identification is set, but _only_ for newly created documents.
            # identificatie is immutable once the document is created
            if identification:
                prop_name = mapper("identificatie")
                prop_type = get_cmis_type(EnkelvoudigInformatieObject, "identificatie")
                props[prop_name] = {"value": str(identification), "type": prop_type}

            # For new documents, the uuid needs to be set
            prop_name = mapper("uuid")
            prop_type = get_cmis_type(EnkelvoudigInformatieObject, "uuid")
            props[prop_name] = {"value": str(uuid.uuid4()), "type": prop_type}

        return props

    def get_document(self, object_id: str) -> "Document":
        """Get latest version of a document with specified objectId

        The objectId contains the specific version in Alfresco

        :param object_id: string, objectId of the document
        :return: Document
        """
        return self.get_content_object(object_id=object_id, object_type=type(self))

    def checkout(self) -> "Document":
        """Checkout a private working copy of the document"""

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            cmis_action="checkOut",
            object_id=str(self.objectId),
        )
        logger.debug(soap_envelope.toprettyxml())

        # FIXME temporary solution due to alfresco raising a 500 AFTER locking the document
        try:
            soap_response = self.request(
                "VersioningService", soap_envelope=soap_envelope.toxml()
            )
            xml_response = extract_xml_from_soap(soap_response)
            logger.debug(pretty_xml(xml_response))
            extracted_data = extract_object_properties_from_xml(
                xml_response, "checkOut"
            )[0]
            pwc_id = extracted_data["properties"]["objectId"]["value"]
        except CmisRuntimeException:
            pwc_document = self.get_latest_version()
            pwc_id = pwc_document.objectId

        return self.get_document(pwc_id)

    def checkin(self, checkin_comment: str, major: bool = True) -> "Document":
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            cmis_action="checkIn",
            object_id=str(self.objectId),
            major=str(major).lower(),
            checkin_comment=checkin_comment,
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "VersioningService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "checkIn")[0]
        doc_id = extracted_data["properties"]["objectId"]["value"]

        return self.get_document(doc_id)

    def get_all_versions(self) -> List["Document"]:
        object_id = self.objectId.split(";")[0]
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            cmis_action="getAllVersions",
            object_id=object_id,
        )
        logger.debug(soap_envelope.toprettyxml())
        soap_response = self.request(
            "VersioningService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))
        extracted_data = extract_object_properties_from_xml(
            xml_response, "getAllVersions"
        )
        return [Document(data) for data in extracted_data]

    def get_private_working_copy(self) -> Union["Document", None]:
        """Get the version of the document with version label 'pwc'"""
        documents = self.get_all_versions()

        for document in documents:
            if document.versionLabel == "pwc":
                return document

    def update_content(self, content: BytesIO, filename: Optional[str] = None):
        self.set_content_stream(content, filename)

    def update_properties(self, properties: dict) -> "Document":

        updated_properties = self._update_properties(properties)
        return self.get_document(updated_properties["properties"]["objectId"]["value"])

    def get_content_stream(self) -> BytesIO:
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=self.objectId,
            cmis_action="getContentStream",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml(), keep_binary=True
        )

        xml_response = extract_xml_from_soap(soap_response, binary=True)
        logger.debug(pretty_xml(xml_response))

        # FIXME find a better way to do this
        return extract_content(soap_response)

    def set_content_stream(self, content: BytesIO, filename: Optional[str] = None):
        content_id = str(uuid.uuid4())
        attachments = [(content_id, content)]

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=self.objectId,
            cmis_action="setContentStream",
            content_id=content_id,
            content_filename=filename,
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
            attachments=attachments,
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

    def delete_object(self):
        """
        Permanently delete the object from the CMIS store, with all its versions.

        By default, all versions should be deleted according to the CMIS standard. If
        the document is currently locked (i.e. there is a private working copy), we need
        to cancel that checkout first.
        """
        latest_version = self.get_latest_version()

        if latest_version.isVersionSeriesCheckedOut:
            soap_envelope = make_soap_envelope(
                auth=(self.user, self.password),
                repository_id=self.main_repo_id,
                object_id=latest_version.objectId,
                cmis_action="cancelCheckOut",
            )
            logger.debug(soap_envelope.toprettyxml())

            soap_response = self.request(
                "VersioningService",
                soap_envelope=soap_envelope.toxml(),
            )
            xml_response = extract_xml_from_soap(soap_response)
            logger.debug(pretty_xml(xml_response))

            refreshed_document = self.get_latest_version()
            return refreshed_document.delete_object()

        return super().delete_object()

    def get_latest_version(self):
        """Get the latest version or the PWC"""

        # This always selects the latest version, and if there is a pwc,
        # Alfresco returns both the pwc and the latest major version, while Corsa only returns the pwc.
        query = CMISQuery("SELECT * FROM drc:document WHERE drc:document__uuid = '%s'")

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=query(self.uuid),
            cmis_action="query",
        )
        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        # Corsa raises an error for queries that return no results
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                error_string = f"Object met objectId '{self.objectId}' bestaat niet in het CMIS connection"
                raise DocumentDoesNotExistError(error_string)
            else:
                raise exc
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))
        extracted_data = extract_object_properties_from_xml(xml_response, "query")
        return extract_latest_version(type(self), extracted_data)


class Gebruiksrechten(CMISContentObject):
    table = "drc:gebruiksrechten"
    name_map = GEBRUIKSRECHTEN_MAP
    type_name = "gebruiksrechten"
    type_class = GebruiksrechtenDoc

    def update_properties(self, properties: dict) -> "Gebruiksrechten":
        """
        Update the properties of an existing gebruiksrechten.

        :param properties: dict, the new properties
        :return: Updated gebruiksrechten
        """
        updated_properties = self._update_properties(properties)

        return self.get_content_object(
            object_id=updated_properties["properties"]["objectId"]["value"],
            object_type=type(self),
        )


class ObjectInformatieObject(CMISContentObject):
    table = "drc:oio"
    name_map = OBJECTINFORMATIEOBJECT_MAP
    type_name = "oio"
    type_class = OioDoc


class Folder(CMISBaseObject):
    table = "cmis:folder"
    type_name = "folder"
    type_class = _Folder

    def get_children_folders(self, child_type: dict = None) -> List:
        """Get all the folders in the current folder

        :param child_type: dict, With keys "value" and "type". The value contains the object type ID of
        the children folders to retrieve.
        """

        if child_type is not None:
            object_type_id = child_type["value"]
            # Alfresco case: the object type ID has an extra prefix (F:drc:zaakfolder, instead of drc:zaakfolder)
            # The prefix needs to be removed for the query
            if len(object_type_id.split(":")) > 2:
                object_type_id = ":".join(object_type_id.split(":")[1:])
        else:
            object_type_id = "cmis:folder"

        query = CMISQuery(f"SELECT * FROM {object_type_id} WHERE cmis:parentId = '%s'")

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=query(str(self.objectId)),
            cmis_action="query",
        )
        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        # Corsa raises an error if the query retrieves 0 results
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                return []
            else:
                raise exc
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "query")
        return [type(self)(folder) for folder in extracted_data]

    def delete_tree(self):
        """Delete the folder and all its contents"""

        # With Corsa, locked documents cause an error, so 'continue_on_failure' is needed
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=self.objectId,
            cmis_action="deleteTree",
            continue_on_failure="true",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))


class ZaakTypeFolder(CMISBaseObject):
    table = "drc:zaaktypefolder"
    name_map = ZAAKTYPE_MAP
    type_name = "zaaktype"
    type_class = ZaakTypeFolderData


class ZaakFolder(CMISBaseObject):
    table = "drc:zaakfolder"
    name_map = ZAAK_MAP
    type_name = "zaak"
    type_class = ZaakFolderData



import datetime
import logging
import mimetypes
import uuid
from datetime import date
from io import BytesIO
from typing import List, Optional, Union

from django.utils import timezone

import pytz

from drc_cmis.utils.mapper import (
    DOCUMENT_MAP,
    GEBRUIKSRECHTEN_MAP,
    OBJECTINFORMATIEOBJECT_MAP,
    ZAAK_MAP,
    ZAAKTYPE_MAP,
    mapper,
)
from drc_cmis.utils.query import CMISQuery
from drc_cmis.utils.utils import extract_latest_version, get_random_string

from .request import CMISRequest

logger = logging.getLogger(__name__)


class CMISBaseObject(CMISRequest):
    name_map = None
    type_name = None
    type_class = None

    def __init__(self, data):
        super().__init__()
        self.data = data

        # Convert any timestamps to datetime objects
        properties = data.get("properties", {})
        for prop_name, prop_details in properties.items():
            if prop_details["type"] == "datetime" and prop_details["value"] is not None:
                prop_details["value"] = timezone.make_aware(
                    datetime.datetime.fromtimestamp(int(prop_details["value"]) / 1000),
                    pytz.timezone(self.time_zone),
                )

        self.properties = properties

    def __getattr__(self, name: str):
        try:
            return super(CMISRequest, self).__getattribute__(name)
        except AttributeError:
            pass

        if name in self.properties:
            return self.properties[name]["value"]

        convert_name = f"cmis:{name}"
        if convert_name in self.properties:
            return self.properties[convert_name]["value"]

        convert_name = f"drc:{name}"
        if self.name_map is not None and name in self.name_map:
            convert_name = self.name_map.get(name)

        if convert_name not in self.properties:
            raise AttributeError(f"No property '{convert_name}'")

        return self.properties[convert_name]["value"]

    @classmethod
    def build_properties(cls, data: dict) -> dict:
        """Construct property dictionary."""

        props = {}
        for key, value in data.items():
            prop_name = mapper(key, type=cls.type_name)
            if not prop_name:
                logger.debug("CMIS_ADAPTER: No property name found for key '%s'", key)
                continue
            if value is not None:
                if isinstance(value, datetime.date) or isinstance(value, datetime.date):
                    value = value.strftime("%Y-%m-%dT%H:%M:%S.000Z")
                props[prop_name] = str(value)

        return props


class CMISContentObject(CMISBaseObject):
    def delete_object(self):
        """Delete all versions of an object"""
        data = {"objectId": self.objectId, "cmisaction": "delete"}
        logger.debug("CMIS_ADAPTER: delete_object: request data: %s", data)
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: delete_object: response data: %s", json_response)
        return json_response

    def get_parent_folders(self) -> List["Folder"]:
        """Get the parent folders of an object.

        An object has multiple parent folders if it has been multifiled.
        """
        params = {
            "objectId": self.objectId,
            "cmisselector": "parents",
        }
        logger.debug("CMIS_ADAPTER: get_parent_folders: request params: %s", params)

        json_response = self.get_request(self.root_folder_url, params=params)
        logger.debug(
            "CMIS_ADAPTER: get_parent_folders: response data: %s", json_response
        )
        return self.get_all_objects(json_response, Folder)

    def move_object(self, target_folder: "Folder"):
        source_folder = self.get_parent_folders()[0]

        data = {
            "objectId": self.objectId,
            "cmisaction": "move",
            "sourceFolderId": source_folder.objectId,
            "targetFolderId": target_folder.objectId,
        }

        logger.debug("CMIS_ADAPTER: move_object: request data: %s", data)

        # invoke the URL
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: move_object: response data: %s", json_response)
        self.data = json_response
        self.properties = json_response.get("properties")
        return self

    def _update_properties(self, properties: dict) -> "CMISContentObject":
        data = {"objectId": self.objectId, "cmisaction": "update"}
        prop_count = 0
        for prop_key, prop_value in properties.items():
            # Skip property because update is not allowed
            if prop_key == "cmis:objectTypeId":
                continue

            if isinstance(prop_value, date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%I:%S.000Z")

            data["propertyId[%s]" % prop_count] = prop_key
            data["propertyValue[%s]" % prop_count] = prop_value
            prop_count += 1
        logger.debug("CMIS_ADAPTER: update_properties: request data: %s", data)

        # invoke the URL
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug(
            "CMIS_ADAPTER: update_properties: response data: %s", json_response
        )
        self.data = json_response
        self.properties = json_response.get("properties")

        return self


class Document(CMISContentObject):
    table = "drc:document"
    name_map = DOCUMENT_MAP

    @classmethod
    def build_properties(
        cls, data: dict, new: bool = True, identification: str = ""
    ) -> dict:

        props = {}
        for key, value in data.items():
            prop_name = mapper(key, type="document")
            if not prop_name:
                logger.debug("CMIS_ADAPTER: No property name found for key '%s'", key)
                continue
            props[prop_name] = value

        # For documents that are not new, the uuid shouldn't be written
        props.pop(mapper("uuid"), None)

        if new:
            # increase likelihood of uniqueness of title by appending a random string
            title, suffix = data.get("titel"), get_random_string()
            if title is not None:
                props["cmis:name"] = f"{title}-{suffix}"

            # make sure the identification is set, but _only_ for newly created documents.
            # identificatie is immutable once the document is created
            if identification:
                prop_name = mapper("identificatie")
                props[prop_name] = identification

            # For new documents, the uuid needs to be set
            props[mapper("uuid", type="document")] = str(uuid.uuid4())

        return props

    def checkout(self):
        data = {"objectId": self.objectId, "cmisaction": "checkOut"}
        logger.debug("CMIS_ADAPTER: checkout: request data: %s", data)
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: checkout: response data: %s", json_response)
        return Document(json_response)

    def update_content(self, content: BytesIO, filename: Optional[str] = None):
        self.set_content_stream(content, filename)

    def update_properties(self, properties: dict):
        return self._update_properties(properties)

    def get_private_working_copy(self) -> Union["Document", None]:
        """
        Retrieve the private working copy version of a document.
        """
        logger.debug(
            "CMIS_ADAPTER: Fetching the PWC for the current document (UUID '%s')",
            self.uuid,
        )

        if self.versionSeriesCheckedOutId is None:
            all_versions = self.get_all_versions()
            for document in all_versions:
                if document.versionLabel == "pwc":
                    return document
        else:
            # http://docs.oasis-open.org/cmis/CMIS/v1.1/os/CMIS-v1.1-os.html#x1-5590004
            params = {
                "cmisselector": "object",  # get the object rather than the content
                "objectId": self.versionSeriesCheckedOutId,
            }
            logger.debug(
                "CMIS_ADAPTER: get_private_working_copy: request params: %s", params
            )

            data = self.get_request(self.root_folder_url, params)
            logger.debug(
                "CMIS_ADAPTER: get_private_working_copy: response data: %s", data
            )
            return type(self)(data)

    def get_latest_version(self):
        """Get the latest version or the PWC"""
        query = CMISQuery("SELECT * FROM drc:document WHERE drc:document__uuid = '%s'")

        data = {
            "cmisaction": "query",
            "statement": query(self.uuid),
        }
        logger.debug("CMIS_ADAPTER: get_latest_version: request data: %s", data)
        json_response = self.post_request(self.base_url, data)
        logger.debug(
            "CMIS_ADAPTER: get_latest_version: response data: %s", json_response
        )

        return extract_latest_version(type(self), json_response.get("results"))

    def checkin(self, checkin_comment, major=True):
        props = {
            "objectId": self.objectId,
            "cmisaction": "checkIn",
            "checkinComment": checkin_comment,
            "major": major,
        }
        logger.debug("CMIS_ADAPTER: checkin: request data: %s", props)

        # invoke the URL
        json_response = self.post_request(self.root_folder_url, props)
        logger.debug("CMIS_ADAPTER: checkin: response data: %s", json_response)
        return Document(json_response)

    def set_content_stream(self, content_file: BytesIO, filename: Optional[str] = None):
        data = {"objectId": self.objectId, "cmisaction": "setContent"}

        mimetype = None
        # need to determine the mime type
        if filename:
            mimetype, _encoding = mimetypes.guess_type(
                filename
            )  # If the extension is not recognised, mimetype is None

        if not mimetype:
            mimetype = "application/binary"

        files = {self.name: (self.name, content_file, mimetype)}
        logger.debug("CMIS_ADAPTER: set_content_stream: request data: %s", data)

        json_response = self.post_request(self.root_folder_url, data=data, files=files)
        logger.debug(
            "CMIS_ADAPTER: set_content_stream: response data: %s", json_response
        )
        return Document(json_response)

    def get_content_stream(self) -> BytesIO:
        params = {"objectId": self.objectId, "cmisaction": "content"}
        logger.debug("CMIS_ADAPTER: get_content_stream: request params: %s", params)
        file_content = self.get_request(self.root_folder_url, params=params)
        logger.debug(
            "CMIS_ADAPTER: get_content_stream: retrieved file length %i",
            len(file_content),
        )
        return BytesIO(file_content)

    def get_all_versions(self) -> List["Document"]:
        """
        Retrieve all versions for a given document.

        Versions are ordered by most-recent first based on cmis:creationDate. If there
        is a PWC, it shall be the first object.

        http://docs.oasis-open.org/cmis/CMIS/v1.1/errata01/os/CMIS-v1.1-errata01-os-complete.html#x1-3440006
        """

        params = {"objectId": self.objectId, "cmisselector": "versions"}
        logger.debug("CMIS_ADAPTER: get_all_versions: request params: %s", params)
        all_versions = self.get_request(self.root_folder_url, params=params)
        logger.debug("CMIS_ADAPTER: get_all_versions: response data: %s", all_versions)
        return [Document(data) for data in all_versions]

    def delete_object(self) -> None:
        """
        Permanently delete the object from the CMIS store, with all its versions.

        By default, all versions should be deleted according to the CMIS standard. If
        the document is currently locked (i.e. there is a private working copy), we need
        to cancel that checkout first.
        """
        latest_version = self.get_latest_version()
        if latest_version.isVersionSeriesCheckedOut:
            cancel_checkout_data = {
                "cmisaction": "cancelCheckout",
                "objectId": latest_version.objectId,
            }
            logger.debug(
                "CMIS_ADAPTER: delete_object: request data: %s", cancel_checkout_data
            )

            response = self.post_request(
                self.root_folder_url, data=cancel_checkout_data
            )

            logger.debug("CMIS_ADAPTER: delete_object: response data: %s", response)

            refreshed_document = self.get_latest_version()
            return refreshed_document.delete_object()
        return super().delete_object()


class Gebruiksrechten(CMISContentObject):
    table = "drc:gebruiksrechten"
    name_map = GEBRUIKSRECHTEN_MAP
    type_name = "gebruiksrechten"

    def update_properties(self, properties: dict) -> "Gebruiksrechten":
        """
        Update the properties of an existing gebruiksrechten.

        :param properties: dict, the new properties
        :return: Updated gebruiksrechten
        """
        return self._update_properties(properties)


class ObjectInformatieObject(CMISContentObject):
    table = "drc:oio"
    name_map = OBJECTINFORMATIEOBJECT_MAP
    type_name = "oio"


class Folder(CMISBaseObject):
    table = "cmis:folder"

    def get_children_folders(self, child_type: Union[str, dict] = None) -> List:
        """Get all the folders in the current folder

        :param child_type: str or dict, Contains the object type ID of the children folders to retrieve.
        If it is a dict, then the child type is the value of the key "value".
        """

        if child_type is not None:
            if isinstance(child_type, dict):
                object_type_id = child_type["value"]
            else:
                object_type_id = child_type

            # Alfresco case: the object type ID has an extra prefix (F:drc:zaakfolder, instead of drc:zaakfolder)
            # The prefix needs to be removed for the query
            if len(object_type_id.split(":")) > 2:
                object_type_id = ":".join(object_type_id.split(":")[1:])
        else:
            object_type_id = "cmis:folder"

        data = {
            "cmisaction": "query",
            "statement": f"SELECT * FROM {object_type_id} WHERE IN_FOLDER('{self.objectId}')",
        }
        logger.debug("CMIS_ADAPTER: get_children_folders: request data: %s", data)
        json_response = self.post_request(self.base_url, data=data)
        logger.debug(
            "CMIS_ADAPTER: get_children_folders: response data: %s", json_response
        )
        return self.get_all_results(json_response, Folder)

    def delete_tree(self, **kwargs):
        data = {"objectId": self.objectId, "cmisaction": "deleteTree"}
        logger.debug("CMIS_ADAPTER: delete_tree: request data: %s", data)
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: delete_tree: response data: %s", json_response)


class ZaakTypeFolder(CMISBaseObject):
    table = "drc:zaaktypefolder"
    name_map = ZAAKTYPE_MAP
    type_name = "zaaktype"


class ZaakFolder(CMISBaseObject):
    table = "drc:zaakfolder"
    name_map = ZAAK_MAP
    type_name = "zaak"



from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal


class Id:
    pass


class Url:
    pass


class QueriableUrl:
    pass


@dataclass
class EnkelvoudigInformatieObject:
    version_label: Decimal
    object_type_id: Id
    name: str
    integriteit_waarde: str
    titel: str
    bestandsnaam: str
    formaat: str
    ondertekening_soort: str
    beschrijving: str
    identificatie: str
    verzenddatum: date
    taal: str
    indicatie_gebruiksrecht: str
    verwijderd: bool
    status: str
    ontvangstdatum: date
    informatieobjecttype: QueriableUrl
    auteur: str
    vertrouwelijkheidaanduiding: str
    begin_registratie: datetime
    ondertekening_datum: date
    bronorganisatie: str
    integriteit_datum: date
    link: Url
    creatiedatum: date
    versie: Decimal
    lock: str
    uuid: str
    integriteit_algoritme: str
    kopie_van: str


@dataclass
class Gebruiksrechten:
    uuid: str
    version_label: Decimal
    object_type_id: Id
    name: str
    einddatum: datetime
    omschrijving_voorwaarden: str
    informatieobject: QueriableUrl
    startdatum: datetime
    kopie_van: str


@dataclass
class Oio:
    uuid: str
    version_label: Decimal
    object_type_id: Id
    name: str
    object_type: str
    besluit: QueriableUrl
    zaak: QueriableUrl
    informatieobject: QueriableUrl


@dataclass
class Folder:
    object_type_id: Id


@dataclass
class ZaakFolderData:
    object_type_id: Id
    url: QueriableUrl
    identificatie: str
    zaaktype: QueriableUrl
    bronorganisatie: str


@dataclass
class ZaakTypeFolderData:
    object_type_id: Id
    url: QueriableUrl
    identificatie: str


CONVERTER = {
    str: "propertyString",
    date: "propertyDateTime",
    datetime: "propertyDateTime",
    Decimal: "propertyDecimal",
    bool: "propertyBoolean",
    Id: "propertyId",
    Url: "propertyString",
    QueriableUrl: "propertyString",
}


def get_type(model: type, name: str) -> type:
    """Return the type of a field"""
    type_annotations = getattr(model, "__annotations__")
    if type_annotations.get(name):
        return type_annotations.get(name)


def get_cmis_type(model: type, name: str) -> str:
    """Get the CMIS type of a property"""
    type_annotations = getattr(model, "__annotations__")
    property_type = type_annotations[name]
    return CONVERTER[property_type]



from decimal import Decimal
from typing import List, TypeVar

from django.utils.crypto import get_random_string as _get_random_string

from drc_cmis.utils.exceptions import DocumentDoesNotExistError

Document = TypeVar("Document")


def get_random_string(number: int = 6) -> str:
    allowed_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return _get_random_string(length=number, allowed_chars=allowed_chars)


def build_query_filters(
    filters: dict,
    object_type: str = None,
    filter_string: str = "",
    strip_end: bool = False,
):
    """Build filters for SQL query"""
    from drc_cmis.utils.mapper import mapper

    if filters:
        for key, value in filters.items():
            if object_type is None:
                if mapper(key):
                    key = mapper(key)
                elif mapper(key, type="connection"):
                    key = mapper(key, type="connection")
                elif mapper(key, type="gebruiksrechten"):
                    key = mapper(key, type="gebruiksrechten")
                elif mapper(key, type="oio"):
                    key = mapper(key, type="oio")
            else:
                key = mapper(key, type=object_type)

            if value and value in ["NULL", "NOT NULL"]:
                filter_string += f"{key} IS {value} AND "
            elif isinstance(value, Decimal):
                filter_string += f"{key} = {value} AND "
            elif isinstance(value, list):
                if len(value) == 0:
                    continue
                filter_string += "( "
                for item in value:
                    sub_filter_string = build_query_filters({key: item}, strip_end=True)
                    filter_string += f"{sub_filter_string} OR "
                filter_string = filter_string[:-3]
                filter_string += " ) AND "
            elif value:
                filter_string += f"{key} = '{value}' AND "

    if strip_end and filter_string[-4:] == "AND ":
        filter_string = filter_string[:-4]

    return filter_string


def extract_latest_version(object_type: type, extracted_data: List) -> Document:
    """Get the latest version

    If there is a private working copy, then the pwc is returned.

    :param object_type: type of the object to be returned
    :param extracted_data: dict, the results of a query
    "SELECT * FROM drc:document WHERE drc:document__uuid = '<uuid>'"
    :return: Document, the latest not locked version of a document
    """
    error_string = "Document bestaat niet in het CMIS connection"
    does_not_exist = DocumentDoesNotExistError(error_string)

    if len(extracted_data) == 0:
        raise does_not_exist
    elif len(extracted_data) == 1:
        return object_type(extracted_data[0])
    elif len(extracted_data) == 2:
        # In this case there is both the latest version and pwc
        # return the latest version
        for doc_data in extracted_data:
            if doc_data["properties"]["cmis:versionLabel"]["value"] == "pwc":
                return object_type(doc_data)



import json
import os

from django.apps import AppConfig
from django.conf import settings

from drc_cmis.utils import mapper
from drc_cmis.utils.mapper import (
    DOCUMENT_MAP,
    GEBRUIKSRECHTEN_MAP,
    OBJECTINFORMATIEOBJECT_MAP,
    ZAAK_MAP,
    ZAAKTYPE_MAP,
)


class CMISConfig(AppConfig):
    name = "drc_cmis"
    verbose_name = "CMIS"
    app_name = "cmis"

    def ready(self):
        maps = {
            "ZAAKTYPE_MAP": ZAAKTYPE_MAP,
            "ZAAK_MAP": ZAAK_MAP,
            "DOCUMENT_MAP": DOCUMENT_MAP,
            "GEBRUIKSRECHTEN_MAP": GEBRUIKSRECHTEN_MAP,
            "OBJECTINFORMATIEOBJECT_MAP": OBJECTINFORMATIEOBJECT_MAP,
        }

        mapper_file_path = settings.CMIS_MAPPER_FILE

        with open(mapper_file_path, "r") as f:
            loaded_cmis_maps = json.load(f)

        for map_name, map_values in loaded_cmis_maps.items():
            for key, value in map_values.items():
                if map_name in maps and key in maps[map_name]:
                    maps[map_name][key] = value  # Add some validation

        errors = []
        for map_name, map_values in maps.items():
            for key, value in map_values.items():
                if maps.get(map_name).get(key) is None:
                    errors.append(f"{key} in {map_name}")

        if len(errors) > 0:
            raise Exception(
                f"Unsuccessful CMIS configuration. "
                f"The following fields in {os.path.basename(mapper_file_path)} are "
                f"unfilled: {', '.join(er for er in errors)}"
            )

        self.refresh_reverse_maps()

    def refresh_reverse_maps(self):
        """
        The maps are now configurable through a json file, so once the maps are initialised,
        the reverse maps need to be updated
        """
        mapper.REVERSE_ZAAKTYPE_MAP = {
            value: key for key, value in ZAAKTYPE_MAP.items()
        }
        mapper.REVERSE_ZAAK_MAP = {value: key for key, value in ZAAK_MAP.items()}
        mapper.REVERSE_DOCUMENT_MAP = {
            value: key for key, value in DOCUMENT_MAP.items()
        }
        mapper.REVERSE_GEBRUIKSRECHTEN_MAP = {
            value: key for key, value in GEBRUIKSRECHTEN_MAP.items()
        }
        mapper.REVERSE_OBJECTINFORMATIEOBJECT_MAP = {
            value: key for key, value in OBJECTINFORMATIEOBJECT_MAP.items()
        }



class TempDocument:
    def __init__(
        self,
        url=None,
        auteur=None,
        bestandsnaam=None,
        creatiedatum=None,
        vertrouwelijkheidaanduiding=None,
        taal=None,
    ):
        self.url = url
        self.auteur = auteur
        self.bestandsnaam = bestandsnaam
        self.creatiedatum = creatiedatum
        self.vertrouwelijkheidaanduiding = vertrouwelijkheidaanduiding
        self.taal = taal


class BaseDRCStorageBackend:
    """
    This is the base Backend storage for the DRC where it should all be based on.
    """

    def get_folder(self, zaak_url):
        raise NotImplementedError()

    def create_folder(self, zaak_url):
        raise NotImplementedError()

    def rename_folder(self, old_zaak_url, new_zaak_url):
        raise NotImplementedError()

    def remove_folder(self, zaak_url):
        raise NotImplementedError()

    def get_document(self, enkelvoudiginformatieobject):
        raise NotImplementedError()

    def create_document(self, enkelvoudiginformatieobject, bestand=None, link=None):
        raise NotImplementedError()

    def update_document(
        self, enkelvoudiginformatieobject, updated_values, bestand=None, link=None
    ):
        raise NotImplementedError()

    def remove_document(self, enkelvoudiginformatieobject):
        raise NotImplementedError()

    def move_document(self, enkelvoudiginformatieobject, zaak_url):
        raise NotImplementedError()



import logging
import mimetypes
import re
import uuid
from datetime import timedelta
from io import BytesIO
from typing import List, Optional, Tuple
from xml.dom import minidom

from django.utils import timezone

from cmislib.util import parsePropValue

from drc_cmis.models import UrlMapping
from drc_cmis.utils.utils import get_random_string

logger = logging.getLogger(__name__)


def extract_object_properties_from_xml(xml_data: str, cmis_action: str) -> List[dict]:
    """Extract properties returned in a XML SOAP response.

    Function adapted from cmislib. It parses a XML response and extracts properties to a dictionary.
    The dictionary has format

    {"properties":
        {"property_name_1": {"value": "property_value_1"}, "property_name_2": {"value": "property_value_2"}}
    }

    so that it is in the same format as when the browser binding is used.
    For each document/folder in the response, a dictionary is created.
    All the dictionaries are then combined to a list.

    :param xml_data: string, XML data
    :param cmis_action: string, name of the CMIS action that was used in the request, e.g. createDocument
    :return: list of dictionaries with the properties.
    """

    def extract_properties(xml_node: minidom.Element) -> dict:
        """Extract properties in the <ns2:properties> tag of the provided node

        :param xml_node: minidom.Element, parsed XML node
        :return: dict, with the extracted properties.
        """
        properties = {}
        for property_node in xml_node.getElementsByTagName("ns2:properties"):
            for child_node in property_node.childNodes:
                try:
                    property_name = child_node.attributes["propertyDefinitionId"].value
                except KeyError:
                    continue

                node_values = child_node.getElementsByTagName("ns2:value")
                if len(node_values) == 0 or len(node_values[0].childNodes) == 0:
                    properties[property_name] = {"value": None}
                else:
                    properties[property_name] = {
                        "value": parsePropValue(
                            node_values[0].childNodes[0].data, child_node.localName
                        )
                    }
        return properties

    parsed_xml = minidom.parseString(xml_data)

    all_objects = []

    # When creating documents and folders this extracts the objectId
    for action_node in parsed_xml.getElementsByTagName(f"{cmis_action}Response"):
        for child_node in action_node.childNodes:
            node_name = child_node.nodeName

            if node_name == "objectId":
                extracted_properties = {
                    node_name: {"value": child_node.firstChild.nodeValue}
                }
                all_objects.append({"properties": extracted_properties})
            if node_name == "object":
                extracted_properties = extract_properties(child_node)
                all_objects.append({"properties": extracted_properties})
            if node_name == "objects" or node_name == "parents":
                if len(child_node.getElementsByTagName("objects")) > 0:
                    for object_nodes in child_node.childNodes:
                        if len(object_nodes.getElementsByTagName("ns2:properties")) > 0:
                            extracted_properties = extract_properties(object_nodes)
                            all_objects.append({"properties": extracted_properties})
                elif len(child_node.getElementsByTagName("ns2:properties")) > 0:
                    extracted_properties = extract_properties(child_node)
                    all_objects.append({"properties": extracted_properties})

    return all_objects


def extract_repository_ids_from_xml(xml_data: str) -> List:
    parsed_xml = minidom.parseString(xml_data)

    repository_id_nodes = parsed_xml.getElementsByTagName("repositoryId")
    return [node.firstChild.nodeValue for node in repository_id_nodes]


def extract_root_folder_id_from_xml(xml_data: str) -> str:
    parsed_xml = minidom.parseString(xml_data)

    folder_id_node = parsed_xml.getElementsByTagName("ns2:rootFolderId")[0]
    return folder_id_node.firstChild.nodeValue


def extract_repo_info_from_xml(xml_data: str) -> dict:
    parsed_xml = minidom.parseString(xml_data)

    properties = {}

    for info_node in parsed_xml.getElementsByTagName("repositoryInfo"):
        for property_node in info_node.childNodes:
            property_name = property_node.nodeName.lstrip("ns2:")
            property_value = property_node.firstChild.nodeValue
            if property_value is not None:
                properties[property_name] = property_value

    return properties


def extract_num_items(xml_data: str) -> int:
    """Extract the number of items in the SOAP XML returned by a query"""
    parsed_xml = minidom.parseString(xml_data)
    folder_id_node = parsed_xml.getElementsByTagName("numItems")[0]
    return int(folder_id_node.firstChild.nodeValue)


def extract_content_stream_properties_from_xml(xml_data: str) -> dict:
    parsed_xml = minidom.parseString(xml_data)

    properties = {}
    for content_stream_node in parsed_xml.getElementsByTagName("contentStream"):
        for prop_node in content_stream_node.childNodes:
            if prop_node.nodeName == "stream":
                value = prop_node.firstChild.attributes["href"].nodeValue
            else:
                value = prop_node.firstChild.nodeValue
            properties[prop_node.nodeName] = value

    return properties


# FIXME find a better way to do this
def extract_content(soap_response_body: bytes) -> BytesIO:
    content = re.search(
        "Content-Disposition: attachment;.+?\r\n\r\n(.+?)\r\n--uuid:.+?--".encode(
            "utf-8"
        ),
        soap_response_body,
        re.DOTALL,
    ).group(1)

    return BytesIO(content)


def make_soap_envelope(
    cmis_action: str,
    auth: Tuple[str, str],
    repository_id: Optional[str] = None,
    properties: Optional[dict] = None,
    statement: Optional[str] = None,
    object_id: Optional[str] = None,
    folder_id: Optional[str] = None,
    content_id: Optional[str] = None,
    content_filename: Optional[str] = None,
    major: Optional[str] = None,
    checkin_comment: Optional[str] = None,
    source_folder_id: Optional[str] = None,
    target_folder_id: Optional[str] = None,
    continue_on_failure: Optional[str] = None,
) -> minidom.Document:
    """Create SOAP envelope from data provided

    :param cmis_action: string, the cmis action to perform
    :param auth: tuple, (username, password) for the DMS
    :param repository_id: ID of the main repository (e.g. 8ca7d93b-2286-44b7-bfce-487211e6e9af)
    :param properties: dictionary, properties of the object to create/update
    :param statement: str, SQL statement used in query requests
    :param object_id: str, ID of the node on which to act (e.g.
        workspace://SpacesStore/2bdd4f3d-851f-499b-99ec-142b82ce3c0d)
    :param folder_id: str, ID of a folder (e.g. needed when creating documents)
    :param content_id: str, ID of the content of a document (as the content will be a MTOM attachment)
    :param content_filename: str, name of the file that will be a MTOM attachment. Includes files extension.
    :param major: str, true or false whether the document being checked in is a major version
    :param checkin_comment: str, comment when checking in a document
    :param source_folder_id: str, folder objectId from which to copy a document
    :param target_folder_id: str, folder objectId to which to copy a document
    :param continue_on_failure: str, whether to continue deleting after an error in the deleteTree call
    :return: minidom document
    """

    xml_doc = minidom.Document()

    # Main soap entry element
    entry_element = xml_doc.createElement("soapenv:Envelope")
    entry_element.setAttribute(
        "xmlns:soapenv", "http://schemas.xmlsoap.org/soap/envelope/"
    )
    entry_element.setAttribute(
        "xmlns:ns", "http://docs.oasis-open.org/ns/cmis/messaging/200908/"
    )
    entry_element.setAttribute(
        "xmlns:ns1", "http://docs.oasis-open.org/ns/cmis/core/200908/"
    )
    xml_doc.appendChild(entry_element)

    # Creates the security header
    header_element = xml_doc.createElement("soapenv:Header")
    security_header = xml_doc.createElement("wsse:Security")
    security_header.setAttribute(
        "xmlns:wsse",
        "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
    )
    security_header.setAttribute(
        "xmlns:wsu",
        "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd",
    )
    security_header_id = uuid.uuid4().hex

    # Username token
    username_token = xml_doc.createElement("wsse:UsernameToken")
    username_token.setAttribute(
        "wsu:Id",
        f"UsernameToken-{security_header_id}",
    )
    username_tag = xml_doc.createElement("wsse:Username")
    username_text = xml_doc.createTextNode(auth[0])
    username_tag.appendChild(username_text)
    username_token.appendChild(username_tag)

    password_tag = xml_doc.createElement("wsse:Password")
    password_tag.setAttribute(
        "Type",
        "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText",
    )
    password_text = xml_doc.createTextNode(auth[1])
    password_tag.appendChild(password_text)
    username_token.appendChild(password_tag)

    security_header.appendChild(username_token)

    header_element.appendChild(security_header)
    entry_element.appendChild(header_element)

    # Time stamp
    time_stamp_tag = xml_doc.createElement("wsu:Timestamp")
    time_stamp_tag.setAttribute(
        "wsu:Id",
        f"TS-{security_header_id}",
    )

    created_tag = xml_doc.createElement("wsu:Created")
    created_text = xml_doc.createTextNode(timezone.now().strftime("%Y-%m-%dT%H:%M:%SZ"))
    created_tag.appendChild(created_text)
    time_stamp_tag.appendChild(created_tag)

    expires_tag = xml_doc.createElement("wsu:Expires")
    expires_text = xml_doc.createTextNode(
        (timezone.now() + timedelta(1)).strftime("%Y-%m-%dT%H:%M:%SZ")
    )
    expires_tag.appendChild(expires_text)
    time_stamp_tag.appendChild(expires_tag)

    security_header.appendChild(time_stamp_tag)

    # Body of the document
    body_element = xml_doc.createElement("soapenv:Body")

    # The name of the next tag is the name of the CMIS action to perform (e.g. createFolder)
    action_element = xml_doc.createElement(f"ns:{cmis_action}")

    # Repository ID
    if repository_id is not None:
        repo_element = xml_doc.createElement("ns:repositoryId")
        repo_text = xml_doc.createTextNode(str(repository_id))
        repo_element.appendChild(repo_text)
        action_element.appendChild(repo_element)

    # All the properties
    if properties is not None:
        properties_element = xml_doc.createElement("ns:properties")
        for prop_name, prop_dict in properties.items():
            property_element = xml_doc.createElement(f"ns1:{prop_dict['type']}")
            property_element.setAttribute("propertyDefinitionId", prop_name)

            value_element = xml_doc.createElement("ns1:value")
            value_text = xml_doc.createTextNode(prop_dict["value"])
            value_element.appendChild(value_text)
            property_element.appendChild(value_element)

            properties_element.appendChild(property_element)
        action_element.appendChild(properties_element)

    # For query requests, there is a SQL statement
    if statement is not None:
        query_element = xml_doc.createElement("ns:statement")
        query_text = xml_doc.createTextNode(statement)
        query_element.appendChild(query_text)
        action_element.appendChild(query_element)

    body_element.appendChild(action_element)

    # Folder ID
    if folder_id is not None:
        folder_element = xml_doc.createElement("ns:folderId")
        folder_text = xml_doc.createTextNode(str(folder_id))
        folder_element.appendChild(folder_text)
        action_element.appendChild(folder_element)

    # ObjectId
    if object_id is not None:
        object_id_element = xml_doc.createElement("ns:objectId")
        object_id_text = xml_doc.createTextNode(str(object_id))
        object_id_element.appendChild(object_id_text)
        action_element.appendChild(object_id_element)

    # File content
    if content_id is not None:
        filename = content_filename or get_random_string()
        mimetype, _encoding = mimetypes.guess_type(filename)

        content_element = xml_doc.createElement("ns:contentStream")
        mimetype_element = xml_doc.createElement("ns:mimeType")
        mimetype_txt = xml_doc.createTextNode(mimetype or "application/octet-stream")
        mimetype_element.appendChild(mimetype_txt)
        content_element.appendChild(mimetype_element)

        stream_element = xml_doc.createElement("ns:stream")
        include_element = xml_doc.createElement("inc:Include")
        include_element.setAttribute(
            "xmlns:inc", "http://www.w3.org/2004/08/xop/include"
        )
        include_element.setAttribute("href", f"cid:{content_id}")
        stream_element.appendChild(include_element)
        content_element.appendChild(stream_element)

        filename_element = xml_doc.createElement("ns:filename")
        filename_text = xml_doc.createTextNode(filename)
        filename_element.appendChild(filename_text)
        content_element.appendChild(filename_element)

        action_element.appendChild(content_element)

    if major is not None:
        major_element = xml_doc.createElement("ns:major")
        major_text = xml_doc.createTextNode(major)
        major_element.appendChild(major_text)
        action_element.appendChild(major_element)

    if checkin_comment is not None:
        comment_element = xml_doc.createElement("ns:checkinComment")
        comment_text = xml_doc.createTextNode(checkin_comment)
        comment_element.appendChild(comment_text)
        action_element.appendChild(comment_element)

    if source_folder_id is not None:
        source_folder_element = xml_doc.createElement("ns:sourceFolderId")
        source_folder_text = xml_doc.createTextNode(source_folder_id)
        source_folder_element.appendChild(source_folder_text)
        action_element.appendChild(source_folder_element)

    if target_folder_id is not None:
        target_folder_element = xml_doc.createElement("ns:targetFolderId")
        target_folder_text = xml_doc.createTextNode(target_folder_id)
        target_folder_element.appendChild(target_folder_text)
        action_element.appendChild(target_folder_element)

    if continue_on_failure is not None:
        continue_element = xml_doc.createElement("ns:continueOnFailure")
        continue_text = xml_doc.createTextNode(continue_on_failure)
        continue_element.appendChild(continue_text)
        action_element.appendChild(continue_element)

    entry_element.appendChild(body_element)

    return xml_doc


def extract_xml_from_soap(soap_response, binary=False):
    soap_envelope_start = "<soap:Envelope"
    soap_envelope_end = "</soap:Envelope>"
    if binary:
        soap_envelope_start = soap_envelope_start.encode("UTF-8")
        soap_envelope_end = soap_envelope_end.encode("UTF-8")

    begin_xml = soap_response.find(soap_envelope_start)
    end_xml = soap_response.find(soap_envelope_end) + len(soap_envelope_end)

    return soap_response[begin_xml:end_xml]


def pretty_xml(xml_envelope: str) -> str:
    dom = minidom.parseString(xml_envelope)
    return dom.toprettyxml()


class NoURLMappingException(Exception):
    pass


class URLTooLongException(Exception):
    pass


def shrink_url(long_url: str) -> str:
    """Replace patterns in the long URL with the shorter one in the mapping"""
    matching_pattern = find_matching_pattern(long_url, "long_pattern")
    mapping = UrlMapping.objects.get(long_pattern=matching_pattern)

    short_url = long_url.replace(mapping.long_pattern, mapping.short_pattern)

    if len(short_url) > 100:
        raise URLTooLongException

    return long_url.replace(mapping.long_pattern, mapping.short_pattern)


def expand_url(short_url: str) -> str:
    """Replace patterns in the short URL with the longer one in the mapping"""

    matching_pattern = find_matching_pattern(short_url, "short_pattern")
    mapping = UrlMapping.objects.get(short_pattern=matching_pattern)

    return short_url.replace(mapping.short_pattern, mapping.long_pattern)


def find_matching_pattern(url: str, field: str = None) -> str:
    from drc_cmis.models import CMISConfig

    if field is None:
        field = "long_pattern"

    config = CMISConfig.get_solo()

    patterns = config.urlmapping_set.values_list(field, flat=True)

    matching_patterns = []
    for pattern in patterns:
        if pattern in url:
            matching_patterns.append(pattern)

    if len(matching_patterns) == 0:
        raise NoURLMappingException
    elif len(matching_patterns) > 1:
        return sorted(matching_patterns, key=len, reverse=True)[0]
    else:
        return matching_patterns[0]



ZAAKTYPE_MAP = {
    "object_type_id": None,
    "url": None,
    "identificatie": None,
}

ZAAK_MAP = {
    "object_type_id": None,
    "url": None,
    "identificatie": None,
    "zaaktype": None,
    "bronorganisatie": None,
}

DOCUMENT_MAP = {
    "object_type_id": None,
    "uuid": None,
    "identificatie": None,
    "bronorganisatie": None,
    "creatiedatum": None,
    "titel": None,
    "vertrouwelijkheidaanduiding": None,
    "auteur": None,
    "status": None,
    "beschrijving": None,
    "ontvangstdatum": None,
    "verzenddatum": None,
    "indicatie_gebruiksrecht": None,
    "ondertekening_soort": None,
    "ondertekening_datum": None,
    "informatieobjecttype": None,
    "formaat": None,
    "taal": None,
    "bestandsnaam": None,
    "bestandsomvang": None,
    "versie": None,
    # "inhoud": "drc:",
    "link": None,
    "integriteit_algoritme": None,
    "integriteit_waarde": None,
    "integriteit_datum": None,
    "verwijderd": None,
    "begin_registratie": None,
    "lock": None,
    "kopie_van": None,
}

GEBRUIKSRECHTEN_MAP = {
    "object_type_id": None,
    "uuid": None,
    "informatieobject": None,
    "omschrijving_voorwaarden": None,
    "startdatum": None,
    "einddatum": None,
    "kopie_van": None,
}

OBJECTINFORMATIEOBJECT_MAP = {
    "object_type_id": None,
    "uuid": None,
    "informatieobject": None,
    "object_type": None,
    "zaak": None,
    "besluit": None,
}

REVERSE_ZAAKTYPE_MAP = {value: key for key, value in ZAAKTYPE_MAP.items()}
REVERSE_ZAAK_MAP = {value: key for key, value in ZAAK_MAP.items()}
REVERSE_DOCUMENT_MAP = {value: key for key, value in DOCUMENT_MAP.items()}
REVERSE_GEBRUIKSRECHTEN_MAP = {value: key for key, value in GEBRUIKSRECHTEN_MAP.items()}
REVERSE_OBJECTINFORMATIEOBJECT_MAP = {
    value: key for key, value in OBJECTINFORMATIEOBJECT_MAP.items()
}


def mapper(drc_name, type="document"):
    if type == "zaaktype":
        return ZAAKTYPE_MAP.get(drc_name, None)
    if type == "zaak":
        return ZAAK_MAP.get(drc_name, None)
    if type == "document":
        return DOCUMENT_MAP.get(drc_name, None)
    if type == "gebruiksrechten":
        return GEBRUIKSRECHTEN_MAP.get(drc_name, None)
    if type == "oio":
        return OBJECTINFORMATIEOBJECT_MAP.get(drc_name, None)
    return None


def reverse_mapper(cmis_name, type="document"):
    if type == "zaaktype":
        return REVERSE_ZAAKTYPE_MAP.get(cmis_name, None)
    if type == "zaak":
        return REVERSE_ZAAK_MAP.get(cmis_name, None)
    if type == "document":
        return REVERSE_DOCUMENT_MAP.get(cmis_name, None)
    if type == "gebruiksrechten":
        return REVERSE_GEBRUIKSRECHTEN_MAP.get(cmis_name, None)
    if type == "oio":
        return REVERSE_OBJECTINFORMATIEOBJECT_MAP.get(cmis_name, None)
    return None



import datetime

from drc_cmis.browser.drc_document import Folder


def create_json_request_body(target_folder: Folder, properties: dict) -> dict:
    """Create the body of the JSON request from the target folder and the document properties

    The target properties are already converted to CMIS names.
    """
    data = {
        "objectId": target_folder.objectId,
        "cmisaction": "createDocument",
        "propertyId[0]": "cmis:name",
        "propertyValue[0]": properties.pop("cmis:name"),
    }

    data["propertyId[1]"] = "cmis:objectTypeId"
    if "cmis:objectTypeId" in properties.keys():
        data["propertyValue[1]"] = properties.pop("cmis:objectTypeId")
    else:
        data["propertyValue[1]"] = "drc:document"

    prop_count = 2
    for prop_key, prop_value in properties.items():
        if isinstance(prop_value, datetime.date) or isinstance(
            prop_value, datetime.datetime
        ):
            prop_value = prop_value.strftime("%Y-%m-%dT%H:%M:%S")

        data[f"propertyId[{prop_count}]"] = prop_key
        data[f"propertyValue[{prop_count}]"] = prop_value
        prop_count += 1

    return data



from urllib.parse import urlparse

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.forms import BaseInlineFormSet
from django.utils.translation import ugettext_lazy as _

from .models import CMISConfig


class UrlMappingInlineFormSet(BaseInlineFormSet):
    def clean(self):
        for form in self.forms:
            if (
                form.is_valid()
                and form.cleaned_data.get("long_pattern")
                and form.cleaned_data.get("short_pattern")
            ):
                parsed_long_url = urlparse(form.cleaned_data["long_pattern"])
                parsed_short_url = urlparse(form.cleaned_data["short_pattern"])

                if parsed_long_url.scheme != parsed_short_url.scheme:
                    raise ValidationError(
                        _(
                            "Use the same scheme (e.g. http or https) for both the long and short pattern."
                        )
                    )


class CMISConfigAdminForm(forms.ModelForm):
    class Meta:
        model = CMISConfig
        fields = "__all__"

    # This method is used from the custom template to decide whether to render the URL mapping formset
    def cmis_url_mapping_enabled(self):
        return settings.CMIS_URL_MAPPING_ENABLED



# Generated by Django 2.2.10 on 2021-01-06 16:02

from django.db import migrations, models

import drc_cmis.validators


class Migration(migrations.Migration):

    dependencies = [
        ("drc_cmis", "0013_auto_20201224_1302"),
    ]

    operations = [
        migrations.AlterField(
            model_name="urlmapping",
            name="long_pattern",
            field=models.CharField(
                help_text="Part of the URL to replace. Example: https://somedomain.nl/somesubpath",
                max_length=1000,
                unique=True,
                validators=[drc_cmis.validators.url_mapping_validator],
                verbose_name="Long pattern",
            ),
        ),
        migrations.AlterField(
            model_name="urlmapping",
            name="short_pattern",
            field=models.CharField(
                help_text="Replacement string to make the URL shorter. Example: https://sd.nl",
                max_length=15,
                unique=True,
                validators=[drc_cmis.validators.url_mapping_validator],
                verbose_name="Short pattern",
            ),
        ),
    ]



import logging
import re
import uuid
from io import BytesIO
from typing import List, Optional, Union
from uuid import UUID

from django.conf import settings
from django.utils.crypto import constant_time_compare

from cmislib.domain import CmisId

from drc_cmis.client import CMISClient
from drc_cmis.utils.exceptions import (
    CmisRuntimeException,
    CmisUpdateConflictException,
    DocumentDoesNotExistError,
    DocumentExistsError,
    DocumentLockedException,
    DocumentNotLockedException,
    FolderDoesNotExistError,
    LockDidNotMatchException,
)
from drc_cmis.utils.mapper import mapper, reverse_mapper
from drc_cmis.utils.query import CMISQuery
from drc_cmis.utils.utils import (
    build_query_filters,
    extract_latest_version,
    get_random_string,
)
from drc_cmis.webservice.data_models import (
    EnkelvoudigInformatieObject,
    Gebruiksrechten as GebruiksRechtDoc,
    Oio,
    QueriableUrl,
    get_cmis_type,
    get_type,
)
from drc_cmis.webservice.drc_document import (
    CMISBaseObject,
    CMISContentObject,
    Document,
    Folder,
    Gebruiksrechten,
    ObjectInformatieObject,
    ZaakFolder,
    ZaakTypeFolder,
)
from drc_cmis.webservice.request import SOAPCMISRequest
from drc_cmis.webservice.utils import (
    extract_object_properties_from_xml,
    extract_repo_info_from_xml,
    extract_xml_from_soap,
    make_soap_envelope,
    pretty_xml,
    shrink_url,
)

logger = logging.getLogger(__name__)


class SOAPCMISClient(CMISClient, SOAPCMISRequest):
    """CMIS client for Web service binding (CMIS 1.0)"""

    document_type = Document
    gebruiksrechten_type = Gebruiksrechten
    oio_type = ObjectInformatieObject
    folder_type = Folder
    zaakfolder_type = ZaakFolder
    zaaktypefolder_type = ZaakTypeFolder

    def get_repository_info(self) -> dict:
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            cmis_action="getRepositoryInfo",
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "RepositoryService", soap_envelope=soap_envelope.toxml()
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        return extract_repo_info_from_xml(xml_response)

    @property
    def vendor(self) -> str:
        repo_info = self.get_repository_info()
        return repo_info["vendorName"]

    def query(
        self, return_type_name: str, lhs: List[str] = None, rhs: List[str] = None
    ) -> List[CMISBaseObject]:
        """Perform an SQL query in the DMS

        :param return_type_name: string, either Folder, Document, Oio or Gebruiksrechten
        :param lhs: list of strings, with the LHS of the SQL query
        :param rhs: list of strings, with the RHS of the SQL query
        :return: type, either Folder, Document, Oio or Gebruiksrechten
        """

        return_type = self.get_return_type(return_type_name)

        processed_rhs = rhs
        # Any query that filters based on URL fields needs to be converted to use the short URL version
        if settings.CMIS_URL_MAPPING_ENABLED and lhs is not None and rhs is not None:
            processed_rhs = []

            # Join all the queries and find which fields are used to filter
            joined_lhs = " ".join(lhs)
            column_names = re.findall(r"([a-z]+?:.+?__[a-z]+)", joined_lhs)

            for index, item_rhs in enumerate(rhs):
                column_name = column_names[index]
                property_name = reverse_mapper(
                    column_name, type=return_type_name.lower()
                )
                if (
                    property_name is not None
                    and get_type(return_type.type_class, property_name) == QueriableUrl
                    and item_rhs != ""
                ):
                    processed_rhs.append(shrink_url(item_rhs))
                else:
                    processed_rhs.append(item_rhs)

        table = return_type.table
        where = (" WHERE " + " AND ".join(lhs)) if lhs else ""
        query = CMISQuery("SELECT * FROM %s%s" % (table, where))
        statement = query(*processed_rhs) if processed_rhs else query()

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=statement,
            cmis_action="query",
        )

        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        # Corsa raises an error if the query retrieves 0 results
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                return []
            else:
                raise exc

        xml_response = extract_xml_from_soap(soap_response)

        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "query")

        return [return_type(cmis_object) for cmis_object in extracted_data]

    def create_folder(self, name: str, parent_id: str, data: dict = None) -> Folder:
        """Create a new folder inside a parent

        :param name: string, name of the new folder to create
        :param parent_id: string, cmis:objectId of the parent folder
        :param data: dict, contains the properties of the folder to create.
            The names of the properties are already converted to cmis names (e.g. drc:zaaktype__url)
        :return: Folder, the created folder
        """

        object_type_id = CmisId("cmis:folder")

        properties = {
            "cmis:objectTypeId": {"value": object_type_id, "type": "propertyId"},
            "cmis:name": {"value": name, "type": "propertyString"},
        }

        if data is not None:
            properties.update(data)

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=parent_id,
            properties=properties,
            cmis_action="createFolder",
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(
            xml_response, "createFolder"
        )[0]

        # Creating a folder only returns the objectId
        folder_id = extracted_data["properties"]["objectId"]["value"]

        return self.get_folder(folder_id)

    def get_folder(self, object_id: str) -> Folder:
        """Retrieve folder with given objectId"""

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=object_id,
            cmis_action="getObject",
        )

        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "ObjectService", soap_envelope=soap_envelope.toxml()
            )
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                error_string = f"Folder met objectId '{object_id}' bestaat niet in het CMIS connection"
                raise FolderDoesNotExistError(error_string)
            else:
                raise exc

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "getObject")[
            0
        ]
        return Folder(extracted_data)

    def copy_document(self, document: Document, destination_folder: Folder) -> Document:
        """Copy document to a folder

        :param document: Document, the document to copy
        :param destination_folder: Folder, the folder in which to place the copied document
        :return: the copied document
        """

        # copy the properties from the source document
        drc_properties = {}
        drc_url_properties = {}
        for property_name, property_details in document.properties.items():
            if (
                "cmis:" not in property_name and property_details["value"] is not None
            ) or property_name == "cmis:objectTypeId":
                drc_property_name = reverse_mapper(property_name, type="document")

                # Urls are handled separately, because they are already in the 'short' form
                if (
                    get_type(EnkelvoudigInformatieObject, drc_property_name)
                    == QueriableUrl
                ):
                    drc_url_properties[property_name] = {
                        "value": property_details["value"],
                        "type": "propertyString",
                    }
                else:
                    drc_properties[drc_property_name] = property_details["value"]

        cmis_properties = Document.build_properties(drc_properties, new=False)

        cmis_properties.update(
            **{
                "cmis:objectTypeId": {
                    "value": document.objectTypeId,
                    "type": "propertyId",
                },
                mapper("titel", type="document"): {
                    "value": f"{document.titel} - copy",
                    "type": "propertyString",
                },
                "drc:kopie_van": {
                    "value": document.uuid,
                    "type": "propertyString",
                },  # Keep tack of where this is copied from.
                "drc:document__uuid": {
                    "value": str(uuid.uuid4()),
                    "type": "propertyString",
                },
                **drc_url_properties,
            }
        )

        # Update the cmis:name to make it more unique
        file_name = f"{document.titel}-{get_random_string()}"
        cmis_properties["cmis:name"] = {"value": file_name, "type": "propertyString"}

        # Create copy document
        content_id = str(uuid.uuid4())
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=destination_folder.objectId,
            properties=cmis_properties,
            cmis_action="createDocument",
            content_id=content_id,
            content_filename=drc_properties.get("bestandsnaam"),
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
            attachments=[(content_id, document.get_content_stream())],
        )

        # Creating the document only returns its ID
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(
            xml_response, "createDocument"
        )[0]
        copy_document_id = extracted_data["properties"]["objectId"]["value"]

        return document.get_document(copy_document_id)

    def copy_gebruiksrechten(
        self, source_object: Gebruiksrechten, destination_folder: Folder
    ) -> Gebruiksrechten:
        """Copy a gebruiksrechten to a folder

        :param source_object: Gebruiksrechten, the gebruiksrechten to copy
        :param destination_folder: Folder, the folder in which to place the copied gebruiksrechten
        :return: the copied object
        """

        # copy the properties from the source document
        drc_properties = {}
        drc_url_properties = {}
        for property_name, property_details in source_object.properties.items():
            if (
                "cmis:" not in property_name and property_details["value"] is not None
            ) or property_name == "cmis:objectTypeId":
                drc_property_name = reverse_mapper(
                    property_name, type="gebruiksrechten"
                )

                # Urls are handled separately, because they are already in the 'short' form
                if get_type(GebruiksRechtDoc, drc_property_name) == QueriableUrl:
                    drc_url_properties[property_name] = {
                        "value": property_details["value"],
                        "type": "propertyString",
                    }
                else:
                    drc_properties[drc_property_name] = property_details["value"]

        cmis_properties = Gebruiksrechten.build_properties(drc_properties)

        cmis_properties.update(
            **{
                "cmis:objectTypeId": {
                    "value": source_object.objectTypeId,
                    "type": "propertyId",
                },
                mapper("kopie_van", type="gebruiksrechten"): {
                    "value": source_object.objectId,
                    "type": "propertyString",  # Keep tack of where this is copied from.
                },
                "cmis:name": {"value": get_random_string(), "type": "propertyString"},
                "drc:gebruiksrechten__uuid": {
                    "value": str(uuid.uuid4()),
                    "type": "propertyString",
                },
                **drc_url_properties,
            }
        )

        # Create copy gebruiksrechten
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=destination_folder.objectId,
            properties=cmis_properties,
            cmis_action="createDocument",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
        )

        # Creating the document only returns its ID
        xml_response = extract_xml_from_soap(soap_response)

        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(
            xml_response, "createDocument"
        )[0]
        copy_gebruiksrechten_id = extracted_data["properties"]["objectId"]["value"]

        # Request all the properties of the newly created object
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=copy_gebruiksrechten_id,
            cmis_action="getObject",
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "getObject")[
            0
        ]

        return Gebruiksrechten(extracted_data)

    def create_content_object(
        self, data: dict, object_type: str, destination_folder: Folder = None
    ) -> CMISContentObject:
        """Create a Gebruiksrechten or a ObjectInformatieObject

        :param data: dict, properties of the object to create
        :param object_type: string, either "gebruiksrechten" or "oio"
        :param destination_folder: Folder, the folder in which to place the object
        :return: Either a Gebruiksrechten or ObjectInformatieObject
        """
        assert object_type in [
            "gebruiksrechten",
            "oio",
        ], "'object_type' can be only 'gebruiksrechten' or 'oio'"

        if object_type == "oio":
            return_type = ObjectInformatieObject
            data_class = Oio
        elif object_type == "gebruiksrechten":
            return_type = Gebruiksrechten
            data_class = GebruiksRechtDoc

        if destination_folder is None:
            other_folder = self.get_or_create_other_folder()
            destination_folder = self.get_or_create_folder("Related data", other_folder)

        properties = return_type.build_properties(data)

        properties.setdefault(
            "cmis:objectTypeId",
            {
                "value": f"{self.get_object_type_id_prefix(object_type)}drc:{object_type}",
                "type": "propertyId",
            },
        )
        properties.setdefault(
            "cmis:name", {"value": get_random_string(), "type": "propertyString"}
        )
        properties.setdefault(
            mapper("uuid", type=object_type),
            {"value": str(uuid.uuid4()), "type": get_cmis_type(data_class, "uuid")},
        )

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=destination_folder.objectId,
            properties=properties,
            cmis_action="createDocument",
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(
            xml_response, "createDocument"
        )[0]
        new_object_id = extracted_data["properties"]["objectId"]["value"]

        # Request all the properties of the newly created object
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=new_object_id,
            cmis_action="getObject",
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "getObject")[
            0
        ]

        return return_type(extracted_data)

    def get_content_object(
        self, drc_uuid: Union[str, UUID], object_type: str
    ) -> CMISContentObject:
        """Get the gebruiksrechten/oio with specified uuid

        :param drc_uuid: string or UUID, the value of drc:oio__uuid or drc:gebruiksrechten__uuid
        :param object_type: string, either "gebruiksrechten" or "oio"
        :return: Either a Gebruiksrechten or ObjectInformatieObject
        """

        assert object_type in [
            "gebruiksrechten",
            "oio",
        ], "'object_type' can be only 'gebruiksrechten' or 'oio'"

        query = CMISQuery("SELECT * FROM drc:%s WHERE drc:%s__uuid = '%s'")

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=query(object_type, object_type, str(drc_uuid)),
            cmis_action="query",
        )

        logger.debug(soap_envelope.toprettyxml())

        error_string = (
            f"{object_type.capitalize()} {object_type} met identificatie drc:{object_type}__uuid {drc_uuid} "
            f"bestaat niet in het CMIS connection"
        )
        does_not_exist = DocumentDoesNotExistError(error_string)

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        # Corsa raises an error if the query retrieves 0 results
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                raise does_not_exist
            else:
                raise exc

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "query")
        if len(extracted_data) == 0:
            raise does_not_exist

        if object_type == "oio":
            return ObjectInformatieObject(extracted_data[0])
        elif object_type == "gebruiksrechten":
            return Gebruiksrechten(extracted_data[0])

    def create_document(
        self,
        identification: str,
        bronorganisatie: str,
        data: dict,
        content: BytesIO = None,
    ) -> Document:
        """Create a custom Document (with the EnkelvoudigInformatieObject properties)

        :param identification: string, the document ``identificatie``
        :param bronorganisatie: string, The identifier of the organisation.
        :param data: dict, the properties of the document
        :param content: BytesIO, the content of the document
        :return: Document, the document created
        """

        self.check_document_exists(identification, bronorganisatie)

        data.setdefault("versie", "1")
        data.setdefault(
            "object_type_id",
            f"{self.get_object_type_id_prefix('document')}drc:document",
        )
        data["bronorganisatie"] = bronorganisatie

        content_id = str(uuid.uuid4())
        if content is None:
            content = BytesIO()

        # Create Document in default folder
        other_folder = self.get_or_create_other_folder()

        properties = Document.build_properties(
            data, new=True, identification=identification
        )

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            folder_id=other_folder.objectId,
            properties=properties,
            cmis_action="createDocument",
            content_id=content_id,
            content_filename=data.get("bestandsnaam"),
        )

        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService",
            soap_envelope=soap_envelope.toxml(),
            attachments=[(content_id, content)],
        )

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        # Creating the document only returns its ID
        extracted_data = extract_object_properties_from_xml(
            xml_response, "createDocument"
        )[0]
        new_document_id = extracted_data["properties"]["objectId"]["value"]

        # Request all the properties of the newly created document
        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            object_id=new_document_id,
            cmis_action="getObject",
        )
        logger.debug(soap_envelope.toprettyxml())

        soap_response = self.request(
            "ObjectService", soap_envelope=soap_envelope.toxml()
        )
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "getObject")[
            0
        ]

        return Document(extracted_data)

    def lock_document(self, drc_uuid: str, lock: str):
        """Lock a EnkelvoudigInformatieObject with given drc:document__uuid

        :param drc_uuid: string, the value of drc:document__uuid
        :param lock: string, value of the lock
        """
        cmis_doc = self.get_document(drc_uuid)

        already_locked = DocumentLockedException(
            "Document was already checked out", code="double_lock"
        )

        try:
            pwc = cmis_doc.checkout()
            if pwc.lock:
                raise already_locked

            # store the lock value on the PWC so we can compare it later
            lock_property = {
                mapper("lock"): {
                    "value": lock,
                    "type": get_cmis_type(EnkelvoudigInformatieObject, "lock"),
                }
            }
            pwc.update_properties(lock_property)
        except CmisUpdateConflictException as exc:
            raise already_locked from exc

    def unlock_document(
        self, drc_uuid: str, lock: str, force: bool = False
    ) -> Document:
        """Unlock a document with given uuid

        :param drc_uuid: string, the value of drc:document__uuid
        :param lock: string, value of the lock
        :param force: bool, whether to force the unlocking
        :return: Document, the unlocked document
        """
        cmis_doc = self.get_document(drc_uuid)

        if not cmis_doc.isVersionSeriesCheckedOut:
            raise DocumentNotLockedException(
                "Document is not checked out and/or locked."
            )

        if constant_time_compare(cmis_doc.lock, lock) or force:
            lock_property = {
                mapper("lock"): {
                    "value": "",
                    "type": get_cmis_type(EnkelvoudigInformatieObject, "lock"),
                }
            }
            cmis_doc.update_properties(lock_property)
            return cmis_doc.checkin("Updated via Documenten API")

        raise LockDidNotMatchException("Lock did not match", code="unlock-failed")

    # FIXME filters are useless because uuid is unique
    def get_document(self, drc_uuid: str, filters: Optional[dict] = None) -> Document:
        """Retrieve a document in the main repository with given uuid (drc:document__uuid)

        If the document series is checked out, it returns the private working copy

        :param drc_uuid: string, value of the cmis property drc:document__uuid
        :param filters: dict, filters to find the document
        :return: Document, latest document version
        """
        error_string = f"Document met drc:document__uuid {drc_uuid} bestaat niet in het CMIS connection"
        does_not_exist = DocumentDoesNotExistError(error_string)

        if drc_uuid is None:
            raise does_not_exist

        # This always selects the latest version, and if there is a pwc,
        # Alfresco returns both the pwc and the latest major version, while Corsa only returns the pwc.
        query = CMISQuery(
            "SELECT * FROM drc:document WHERE drc:document__uuid = '%s' %s"
        )

        filter_string = build_query_filters(
            filters, filter_string="AND ", strip_end=True
        )

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=query(drc_uuid, filter_string),
            cmis_action="query",
        )
        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        # Corsa raises an error if the query retrieves 0 results
        except CmisRuntimeException as exc:
            if "objectNotFound" in exc.message:
                raise does_not_exist
            else:
                raise exc
        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "query")
        return extract_latest_version(self.document_type, extracted_data)

    def check_document_exists(
        self, identification: Union[str, UUID], bronorganisatie: str
    ) -> None:
        """Check if a document with the same (identificatie, bronorganisatie) already exists in the repository

        :param identification: string, document ``identificatie``
        :param bronorganisatie: string, document ``bronorganisatie``
        """
        cmis_identificatie = mapper("identificatie", type="document")
        cmis_bronorganisatie = mapper("bronorganisatie", type="document")

        query = CMISQuery(
            f"SELECT * FROM drc:document WHERE {cmis_identificatie} = '%s' AND {cmis_bronorganisatie} = '%s'"
        )

        soap_envelope = make_soap_envelope(
            auth=(self.user, self.password),
            repository_id=self.main_repo_id,
            statement=query(str(identification), bronorganisatie),
            cmis_action="query",
        )
        logger.debug(soap_envelope.toprettyxml())

        try:
            soap_response = self.request(
                "DiscoveryService", soap_envelope=soap_envelope.toxml()
            )
        except CmisRuntimeException as exc:
            # Corsa raises an error if the query gives no results, while Alfresco a 200
            if "objectNotFound" in exc.message:
                return
            else:
                raise exc

        xml_response = extract_xml_from_soap(soap_response)
        logger.debug(pretty_xml(xml_response))

        extracted_data = extract_object_properties_from_xml(xml_response, "query")

        if len(extracted_data) > 0:
            raise DocumentExistsError(
                "Een document met dezelfde identificatie en bronorganisatie al bestaat."
            )



# Generated by Django 2.0.13 on 2019-05-23 10:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("drc_cmis", "0001_initial_new")]

    operations = [
        migrations.CreateModel(
            name="CMISFolderLocation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("location", models.CharField(default="", max_length=200)),
            ],
        ),
        migrations.AddField(
            model_name="cmisconfig",
            name="locations",
            field=models.ManyToManyField(to="drc_cmis.CMISFolderLocation"),
        ),
    ]



import logging
from typing import BinaryIO, List, Optional, Tuple, Union

import requests

from drc_cmis.utils.exceptions import (
    CmisBaseException,
    CmisInvalidArgumentException,
    CmisNotSupportedException,
    CmisObjectNotFoundException,
    CmisPermissionDeniedException,
    CmisRepositoryDoesNotExist,
    CmisRuntimeException,
    CmisUpdateConflictException,
)
from drc_cmis.webservice.utils import (
    extract_repository_ids_from_xml,
    extract_root_folder_id_from_xml,
    extract_xml_from_soap,
    make_soap_envelope,
    pretty_xml,
)

logger = logging.getLogger(__name__)


class SOAPCMISRequest:
    _boundary = "------=_Part_52_1132425564.1594208078802"

    _headers = {
        "Content-Type": 'multipart/related; type="application/xop+xml"; start="<rootpart@soapui.org>"; '
        'start-info="application/soap+xml"; boundary="----=_Part_52_1132425564.1594208078802"',
        "SOAPAction": "",
        "MIME-Version": "1.0",
    }

    _envelope_headers = {
        "Content-Type": 'application/xop+xml; charset=UTF-8; type="application/soap+xml"',
        "Content-Transfer-Encoding": "8bit",
        "Content-ID": "<rootpart@soapui.org>",
    }

    _main_repo_id = None

    @property
    def config(self):
        """
        Lazily load the config so that no DB queries are done while Django is starting.
        """
        from drc_cmis.models import CMISConfig

        return CMISConfig.get_solo()

    @property
    def user(self):
        return self.config.client_user

    @property
    def password(self):
        return self.config.client_password

    @property
    def base_url(self):
        """Return the base URL

        For example, for Alfresco running locally the base URL for SOAP requests is
        http://localhost:8082/alfresco/cmisws
        """
        return self.config.client_url

    @property
    def main_repo_id(self) -> str:
        """Get ID of the CMS main repository"""
        return self.get_main_repo_id()

    def get_main_repo_id(self, cache: bool = True) -> str:
        configured_main_repo_id = self.config.main_repo_id
        if configured_main_repo_id and cache:
            return configured_main_repo_id

        if self._main_repo_id is None:
            # Retrieving the IDs of all repositories in the CMS
            soap_envelope = make_soap_envelope(
                auth=(self.user, self.password), cmis_action="getRepositories"
            )

            logger.debug(soap_envelope.toprettyxml())

            soap_response = self.request(
                "RepositoryService", soap_envelope=soap_envelope.toxml()
            )

            xml_response = extract_xml_from_soap(soap_response)
            logger.debug(pretty_xml(xml_response))

            all_repositories_ids = extract_repository_ids_from_xml(xml_response)

            # If no main repository ID is configured, take the ID of the first repository returned.
            if configured_main_repo_id == "":
                self._main_repo_id = all_repositories_ids[0]
            else:
                if configured_main_repo_id not in all_repositories_ids:
                    raise CmisRepositoryDoesNotExist(
                        "The configured repository ID does not exist."
                    )

                self._main_repo_id = configured_main_repo_id

        return self._main_repo_id

    @property
    def root_folder_id(self) -> str:
        """Get the ID of the folder where all folders/documents will be created"""

        if self._root_folder_id is None:
            soap_envelope = make_soap_envelope(
                auth=(self.user, self.password),
                cmis_action="getRepositoryInfo",
                repository_id=self.main_repo_id,
            )
            logger.debug(soap_envelope.toprettyxml())
            soap_response = self.request(
                "RepositoryService", soap_envelope=soap_envelope.toxml()
            )

            xml_response = extract_xml_from_soap(soap_response)
            logger.debug(pretty_xml(xml_response))
            self._root_folder_id = extract_root_folder_id_from_xml(xml_response)

        return self._root_folder_id

    def request(
        self,
        path: str,
        soap_envelope: str,
        attachments: Optional[List[Tuple[str, BinaryIO]]] = None,
        keep_binary: bool = False,
    ) -> Union[str, bytes]:
        """Make request with MTOM attachment.

        :param path: string, path where to post the request
        :param soap_envelope: string, XML which can contain zero or more references to attachments
        (in the form of `cid:<contentId>`)
        :param attachments: list of tuples, each tuple contains the content ID used in the XML (string) and the I/O
        stream for the attachment.
        :param keep_binary: whether to keep the body of the response as binary or convert it to a string.
        :return: string or bytes, the content of the response
        """
        url = f"{self.base_url}/{path.lstrip('/')}"

        envelope_header = ""
        for key, value in self._envelope_headers.items():
            envelope_header += f"{key}: {value}\n"

        # Format the body of the request
        body = f"\n{self._boundary}\n{envelope_header}\n{soap_envelope}\n\n".encode(
            "utf-8"
        )

        # Adding the attachments
        if attachments is not None:
            for attachment in attachments:
                content_id, content_stream = attachment
                file_attachment_headers = {
                    "Content-Type": "application/octet-stream",
                    "Content-Transfer-Encoding": "binary",
                    "Content-ID": f"<{content_id}>",
                }

                xml_attachment_header = ""
                for key, value in file_attachment_headers.items():
                    xml_attachment_header += f"{key}: {value}\n"

                content_stream.seek(0)
                body += f"{self._boundary}\n{xml_attachment_header}\n".encode("utf-8")
                body += content_stream.read()  # Reads binary

        body += f"{self._boundary}--\n".encode("utf-8")
        soap_response = requests.post(url, data=body, headers=self._headers, files=[])
        if not soap_response.ok:
            error = soap_response.text
            if soap_response.status_code == 401:
                raise CmisPermissionDeniedException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=401,
                )
            elif soap_response.status_code == 400:
                raise CmisInvalidArgumentException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=400,
                )
            elif soap_response.status_code == 404:
                raise CmisObjectNotFoundException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=404,
                )
            elif soap_response.status_code == 403:
                raise CmisPermissionDeniedException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=403,
                )
            elif soap_response.status_code == 405:
                raise CmisNotSupportedException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=405,
                )
            elif soap_response.status_code == 409:
                raise CmisUpdateConflictException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=409,
                )
            elif soap_response.status_code == 500:
                raise CmisRuntimeException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=500,
                )
            else:
                raise CmisBaseException(
                    status=soap_response.status_code,
                    url=url,
                    message=error,
                    code=soap_response.status_code,
                )

        if keep_binary:
            return soap_response.content
        return soap_response.content.decode("utf-8")



import datetime
import logging
import uuid
from io import BytesIO
from typing import List, Optional, Union
from uuid import UUID

from django.utils.crypto import constant_time_compare

from drc_cmis.browser.drc_document import (
    CMISContentObject,
    Document,
    Folder,
    Gebruiksrechten,
    ObjectInformatieObject,
    ZaakFolder,
    ZaakTypeFolder,
)
from drc_cmis.browser.request import CMISRequest
from drc_cmis.browser.utils import create_json_request_body
from drc_cmis.client import CMISClient
from drc_cmis.utils.exceptions import (
    CmisInvalidArgumentException,
    CmisUpdateConflictException,
    DocumentDoesNotExistError,
    DocumentExistsError,
    DocumentLockedException,
    FolderDoesNotExistError,
    GetFirstException,
    LockDidNotMatchException,
)
from drc_cmis.utils.mapper import mapper
from drc_cmis.utils.query import CMISQuery
from drc_cmis.utils.utils import (
    build_query_filters,
    extract_latest_version,
    get_random_string,
)

logger = logging.getLogger(__name__)


class CMISDRCClient(CMISClient, CMISRequest):
    """CMIS client for Browser binding (CMIS 1.1)"""

    document_type = Document
    gebruiksrechten_type = Gebruiksrechten
    oio_type = ObjectInformatieObject
    folder_type = Folder
    zaakfolder_type = ZaakFolder
    zaaktypefolder_type = ZaakTypeFolder

    @property
    def root_folder_id(self) -> str:
        """Returns the objectId of the root folder"""
        if self._root_folder_id is None:
            repository_info = self.get_request(self.base_url)
            self._root_folder_id = repository_info["-default-"]["rootFolderId"]

        return self._root_folder_id

    @property
    def vendor(self) -> str:
        repo_info = self.get_repository_info()
        return repo_info["vendorName"]

    def get_repository_info(self) -> dict:
        logger.debug(
            "CMIS_ADAPTER: get_repository_info: GET request url: %s", self.base_url
        )
        response = self.get_request(self.base_url)["-default-"]
        logger.debug("CMIS_ADAPTER: get_repository_info: response: %s", response)
        return response

    # generic querying
    def query(
        self, return_type_name: str, lhs: List[str] = None, rhs: List[str] = None
    ):
        return_type = self.get_return_type(return_type_name)
        table = return_type.table
        where = (" WHERE " + " AND ".join(lhs)) if lhs else ""
        query = CMISQuery("SELECT * FROM %s%s" % (table, where))
        statement = query(*rhs) if rhs else query()

        body = {"cmisaction": "query", "statement": statement}
        logger.debug("CMIS_ADAPTER: query: request data: %s", body)
        response = self.post_request(self.base_url, body)
        logger.debug("CMIS_ADAPTER: query: response: %s", response)

        return self.get_all_results(response, return_type)

    def create_folder(self, name: str, parent_id: str, properties: dict = None):
        data = {
            "objectId": parent_id,
            "cmisaction": "createFolder",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": name,
            "propertyId[1]": "cmis:objectTypeId",
            "propertyValue[1]": "cmis:folder",
        }

        if properties is not None:
            prop_count = 2
            for prop, value in properties.items():
                data[f"propertyId[{prop_count}]"] = prop
                data[f"propertyValue[{prop_count}]"] = value
                prop_count += 1

        logger.debug("CMIS_ADAPTER: create_folder: request data: %s", data)

        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: create_folder: response data: %s", json_response)

        return Folder(json_response)

    def get_folder(self, object_id: str) -> Folder:
        """Retrieve folder with objectId given"""

        query = CMISQuery("SELECT * FROM cmis:folder WHERE cmis:objectId = '%s'")

        body = {"cmisaction": "query", "statement": query(object_id)}
        logger.debug("CMIS_ADAPTER: get_folder: request data: %s", body)
        json_response = self.post_request(self.base_url, body)
        logger.debug("CMIS_ADAPTER: get_folder: response data: %s", json_response)

        try:
            return self.get_first_result(json_response, Folder)
        except GetFirstException:
            error_string = (
                f"Folder met objectId '{object_id}' bestaat niet in het CMIS connection"
            )
            raise FolderDoesNotExistError(error_string)

    def copy_gebruiksrechten(
        self, source_object: Gebruiksrechten, destination_folder: Folder
    ) -> Gebruiksrechten:
        """Copy a gebruiksrechten to a folder

        :param source_object: Gebruiksrechten, the gebruiksrechten to copy
        :param destination_folder: Folder, the folder in which to place the copied gebruiksrechten
        :return: the copied object
        """
        # copy the properties from the source object
        properties = {
            property_name: property_details["value"]
            for property_name, property_details in source_object.properties.items()
            if "cmis:" not in property_name
        }

        file_name = get_random_string()

        properties.update(
            **{
                "cmis:objectTypeId": source_object.objectTypeId,
                "cmis:name": file_name,
                mapper(
                    "kopie_van", type="gebruiksrechten"
                ): source_object.objectId,  # Keep tack of where this is copied from.
                mapper("uuid", type="gebruiksrechten"): str(uuid.uuid4()),
            }
        )

        data = create_json_request_body(destination_folder, properties)
        logger.debug("CMIS_ADAPTER: copy_gebruiksrechten: request data: %s", data)

        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug(
            "CMIS_ADAPTER: copy_gebruiksrechten: response data: %s", json_response
        )

        return Gebruiksrechten(json_response)

    def copy_document(self, document: Document, destination_folder: Folder) -> Document:
        """Copy document to a folder

        :param document: Document, the document to copy
        :param destination_folder: Folder, the folder in which to place the copied document
        :return: the copied document
        """

        # copy the properties from the source document
        properties = {
            property_name: property_details["value"]
            for property_name, property_details in document.properties.items()
            if "cmis:" not in property_name
        }

        properties.update(
            **{
                "cmis:objectTypeId": document.objectTypeId,
                mapper("titel", type="document"): f"{document.titel} - copy",
                "drc:kopie_van": document.uuid,  # Keep tack of where this is copied from.
                "drc:document__uuid": str(uuid.uuid4()),
            }
        )

        # Update the cmis:name to make it more unique
        file_name = f"{document.titel}-{get_random_string()}"
        properties["cmis:name"] = file_name

        data = create_json_request_body(destination_folder, properties)
        logger.debug("CMIS_ADAPTER: copy_document: request data: %s", data)

        content = document.get_content_stream()
        json_response = self.post_request(self.root_folder_url, data=data)
        logger.debug("CMIS_ADAPTER: copy_document: response data: %s", json_response)

        cmis_doc = Document(json_response)
        content.seek(0)

        return cmis_doc.set_content_stream(content, filename=document.bestandsnaam)

    def create_content_object(
        self, data: dict, object_type: str, destination_folder: Folder = None
    ) -> CMISContentObject:
        """Create a Gebruiksrechten or a ObjectInformatieObject

        :param data: dict, properties of the object to create
        :param object_type: string, either "gebruiksrechten" or "oio"
        :param destination_folder: Folder, a folder where to create the object. If not provided,
            the object will be placed in a temporary folder.
        :return: Either a Gebruiksrechten or ObjectInformatieObject
        """
        assert object_type in [
            "gebruiksrechten",
            "oio",
        ], "'object_type' can be only 'gebruiksrechten' or 'oio'"

        if destination_folder is None:
            other_folder = self.get_or_create_other_folder()
            destination_folder = self.get_or_create_folder("Related data", other_folder)

        properties = {
            mapper(key, type=object_type): value
            for key, value in data.items()
            if mapper(key, type=object_type)
        }

        json_data = {
            "objectId": destination_folder.objectId,
            "cmisaction": "createDocument",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": get_random_string(),
            "propertyId[1]": f"drc:{object_type}__uuid",
            "propertyValue[1]": str(uuid.uuid4()),
        }

        json_data["propertyId[2]"] = "cmis:objectTypeId"
        if "cmis:objectTypeId" in properties.keys():
            json_data["propertyValue[2]"] = properties.pop("cmis:objectTypeId")
        else:
            json_data[
                "propertyValue[2]"
            ] = f"{self.get_object_type_id_prefix(object_type)}drc:{object_type}"

        prop_count = 3
        for prop_key, prop_value in properties.items():
            if isinstance(prop_value, datetime.date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%M:%S.000Z")

            json_data[f"propertyId[{prop_count}]"] = prop_key
            json_data[f"propertyValue[{prop_count}]"] = prop_value
            prop_count += 1

        logger.debug("CMIS_ADAPTER: create_content_object: request data: %s", json_data)
        json_response = self.post_request(self.root_folder_url, data=json_data)
        logger.debug(
            "CMIS_ADAPTER: create_content_object: response data: %s", json_response
        )

        if object_type == "gebruiksrechten":
            return Gebruiksrechten(json_response)
        elif object_type == "oio":
            return ObjectInformatieObject(json_response)

    def get_content_object(
        self, drc_uuid: Union[str, UUID], object_type: str
    ) -> CMISContentObject:
        """Get the gebruiksrechten/oio with specified uuid

        :param drc_uuid: string or UUID, the value of drc:oio__uuid or drc:gebruiksrechten__uuid
        :param object_type: string, either "gebruiksrechten" or "oio"
        :return: Either a Gebruiksrechten or ObjectInformatieObject
        """

        assert object_type in [
            "gebruiksrechten",
            "oio",
        ], "'object_type' can be only 'gebruiksrechten' or 'oio'"

        query = CMISQuery("SELECT * FROM drc:%s WHERE drc:%s__uuid = '%s'")

        data = {
            "cmisaction": "query",
            "statement": query(object_type, object_type, str(drc_uuid)),
        }
        logger.debug("CMIS_ADAPTER: get_content_object: request data: %s", data)

        json_response = self.post_request(self.base_url, data)
        logger.debug(
            "CMIS_ADAPTER: get_content_object: response data: %s", json_response
        )

        try:
            return self.get_first_result(
                json_response, self.get_return_type(object_type)
            )
        except GetFirstException:
            object_title = object_type.capitalize()
            error_string = f"{object_title} met uuid {drc_uuid} bestaat niet in het CMIS connection"
            raise DocumentDoesNotExistError(error_string)

    def create_document(
        self,
        identification: str,
        bronorganisatie: str,
        data: dict,
        content: BytesIO = None,
    ) -> Document:
        """Create a cmis document.

        :param identification: string, A unique identifier for the document.
        :param bronorganisatie: string, The identifier of the organisation.
        :param data: dict, A dict with all the data that needs to be saved on the document.
        :param content: BytesIO, The content of the document.
        :return: document
        """
        self.check_document_exists(identification, bronorganisatie)

        data.setdefault("versie", 1)
        data.setdefault(
            "object_type_id",
            f"{self.get_object_type_id_prefix('document')}drc:document",
        )
        data["bronorganisatie"] = bronorganisatie

        if content is None:
            content = BytesIO()

        # Create Document in default folder
        other_folder = self.get_or_create_other_folder()

        properties = Document.build_properties(
            data, new=True, identification=identification
        )

        json_data = create_json_request_body(other_folder, properties)
        logger.debug("CMIS_ADAPTER: create_document: request data: %s", json_data)

        json_response = self.post_request(self.root_folder_url, data=json_data)
        logger.debug("CMIS_ADAPTER: create_document: response data: %s", json_response)
        cmis_doc = Document(json_response)
        content.seek(0)
        return cmis_doc.set_content_stream(content, filename=data.get("bestandsnaam"))

    def lock_document(self, drc_uuid: str, lock: str):
        """
        Check out the CMIS document and store the lock value for check in/unlock.
        """
        cmis_doc = self.get_document(drc_uuid)

        already_locked = DocumentLockedException(
            "Document was already checked out", code="double_lock"
        )

        try:
            pwc = cmis_doc.checkout()
        except CmisInvalidArgumentException:
            raise already_locked

        if pwc.lock:
            raise already_locked

        try:
            # store the lock value on the PWC so we can compare it later
            pwc.update_properties({mapper("lock"): lock})
        except CmisUpdateConflictException as exc:
            raise already_locked from exc

    def unlock_document(
        self, drc_uuid: str, lock: str, force: bool = False
    ) -> Document:
        """Unlock a document with objectId workspace://SpacesStore/<uuid>"""
        cmis_doc = self.get_document(drc_uuid)
        pwc = cmis_doc.get_private_working_copy()

        if constant_time_compare(pwc.lock, lock) or force:
            pwc.update_properties({mapper("lock"): ""})
            new_doc = pwc.checkin("Updated via Documenten API")
            return new_doc

        raise LockDidNotMatchException("Lock did not match", code="unlock-failed")

    def get_document(
        self, drc_uuid: Optional[str], via_identification=None, filters=None
    ):
        """
        Given a cmis document instance.

        :param drc_uuid: str, the drc:document__uuid
        :return: :class:`AtomPubDocument` object, the latest version of this document
        """
        assert (
            not via_identification
        ), "Support for 'via_identification' is being dropped"

        error_string = f"Document met drc:document__uuid {drc_uuid} bestaat niet in het CMIS connection"
        does_not_exist = DocumentDoesNotExistError(error_string)

        # shortcut - no reason in going over the wire
        if uuid is None:
            raise does_not_exist

        # this always selects the latest version, and if there is a pwc, also the pwc is returned
        query = CMISQuery(
            "SELECT * FROM drc:document WHERE drc:document__uuid = '%s' %s"
        )

        filter_string = build_query_filters(
            filters, filter_string="AND ", strip_end=True
        )
        data = {
            "cmisaction": "query",
            "statement": query(drc_uuid, filter_string),
        }
        logger.debug("CMIS_ADAPTER: get_document: request data: %s", data)
        json_response = self.post_request(self.base_url, data)
        logger.debug("CMIS_ADAPTER: get_document: response data: %s", json_response)

        return extract_latest_version(self.document_type, json_response.get("results"))

    def check_document_exists(
        self, identification: Union[str, UUID], bronorganisatie: str
    ):
        """Check if a document with the same (identificatie, bronorganisatie) already exists in the repository"""

        cmis_identificatie = mapper("identificatie", type="document")
        cmis_bronorganisatie = mapper("bronorganisatie", type="document")

        query = CMISQuery(
            f"SELECT * FROM drc:document WHERE {cmis_identificatie} = '%s' AND {cmis_bronorganisatie} = '%s'"
        )

        data = {
            "cmisaction": "query",
            "statement": query(str(identification), bronorganisatie),
        }
        logger.debug("CMIS_ADAPTER: check_document_exists: request data: %s", data)
        json_response = self.post_request(self.base_url, data)
        logger.debug(
            "CMIS_ADAPTER: check_document_exists: response data: %s", json_response
        )
        if json_response["numItems"] > 0:
            raise DocumentExistsError(
                "Een document met dezelfde identificatie en bronorganisatie al bestaat."
            )



#!/usr/bin/env python
import os
import sys

from django.core.management import execute_from_command_line

if __name__ == "__main__":
    os.environ["DJANGO_SETTINGS_MODULE"] = "settings"

    execute_from_command_line(sys.argv)



# Generated by Django 2.2 on 2019-04-19 08:29

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="EnkelvoudigInformatieObject",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4, unique=True)),
                ("identificatie", models.CharField(default=uuid.uuid4, max_length=40)),
                ("bronorganisatie", models.CharField(max_length=9)),
                ("creatiedatum", models.DateField()),
                ("titel", models.CharField(max_length=200)),
                ("vertrouwelijkheidaanduiding", models.CharField(max_length=200)),
                ("auteur", models.CharField(max_length=200)),
                ("status", models.CharField(max_length=20)),
                ("beschrijving", models.TextField(max_length=1000)),
                ("ontvangstdatum", models.DateField(blank=True, null=True)),
                ("verzenddatum", models.DateField(blank=True, null=True)),
                ("indicatie_gebruiksrecht", models.NullBooleanField(default=None)),
                ("ondertekening_soort", models.CharField(blank=True, max_length=10)),
                ("ondertekening_datum", models.DateField(blank=True, null=True)),
                ("informatieobjecttype", models.URLField()),
                ("formaat", models.CharField(blank=True, max_length=255)),
                ("taal", models.CharField(blank=True, max_length=255)),
                ("bestandsnaam", models.CharField(blank=True, max_length=255)),
                ("inhoud", models.FileField(upload_to="uploads/%Y/%m/")),
                ("link", models.URLField(blank=True)),
                ("integriteit_algoritme", models.CharField(blank=True, max_length=20)),
                ("integriteit_waarde", models.CharField(blank=True, max_length=128)),
                ("integriteit_datum", models.DateField(blank=True, null=True)),
            ],
        ),
    ]



# Generated by Django 2.2.10 on 2020-07-23 13:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("drc_cmis", "0004_auto_20200616_0919"),
    ]

    operations = [
        migrations.RenameField(
            model_name="cmisconfig",
            old_name="base_folder",
            new_name="base_folder_name",
        ),
    ]



import os

from .environ import config

PROJECT_DIR = os.path.dirname(__file__)
BASE_DIR = PROJECT_DIR  # setting present in new startproject

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

SITE_ID = 1

IS_HTTPS = False

USE_TZ = True
TIME_ZONE = "UTC"

INSTALLED_APPS = [
    "vng_api_common",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "solo",
    "drc_cmis",
    "test_app.app",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "test_app.urls"

DEBUG = True
SECRET_KEY = "this-is-really-not-a-secret"
ALLOWED_HOSTS = ["localhost", "127.0.0.1"]

STATIC_URL = "/static/"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(PROJECT_DIR, "database.db"),
    },
}

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [
            os.path.join(PROJECT_DIR, "templates"),
        ],
        "OPTIONS": {
            "debug": DEBUG,
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.media",
            ],
        },
    },
]

CMIS_MAPPER_FILE = os.path.join(PROJECT_DIR, "cmis_mapper.json")
CMIS_URL_MAPPING_ENABLED = config("CMIS_URL_MAPPING_ENABLED", default=False)



class DMSException(Exception):
    def __init__(self, message, *args, **kwargs):
        self.message = message


class FolderDoesNotExistError(DMSException):
    pass


class DocumentExistsError(DMSException):
    pass


class DocumentDoesNotExistError(DMSException):
    pass


class DocumentConflictException(DMSException):
    pass


class DocumentLockConflictException(DMSException):
    pass


class DocumentNotLockedException(DMSException):
    pass


class DocumentLockedException(DMSException):
    pass


class LockDidNotMatchException(DMSException):
    pass


class GetFirstException(Exception):
    pass


class CmisBaseException(Exception):
    """ Common base class for all exceptions. """

    def __init__(self, status, url, message, code):
        self.display_message = f"Error {status}: {code} at {url} detail: {message}"
        super().__init__(self, self.display_message)
        self.status = status
        self.url = url
        self.message = message
        self.code = code


class CmisNoValidResponse(CmisBaseException):
    pass


class CmisInvalidArgumentException(CmisBaseException):
    """ InvalidArgumentException """

    pass


class CmisObjectNotFoundException(CmisBaseException):
    """ ObjectNotFoundException """

    pass


class CmisNotSupportedException(CmisBaseException):
    """ NotSupportedException """

    pass


class CmisPermissionDeniedException(CmisBaseException):
    """ PermissionDeniedException """

    pass


class CmisRuntimeException(CmisBaseException):
    """ RuntimeException """

    pass


class CmisConstraintException(CmisBaseException):
    """ ConstraintException """

    pass


class CmisContentAlreadyExistsException(CmisBaseException):
    """ContentAlreadyExistsException """

    pass


class CmisFilterNotValidException(CmisBaseException):
    """FilterNotValidException """

    pass


class CmisNameConstraintViolationException(CmisBaseException):
    """NameConstraintViolationException """

    pass


class CmisStorageException(CmisBaseException):
    """StorageException """

    pass


class CmisStreamNotSupportedException(CmisBaseException):
    """ StreamNotSupportedException """

    pass


class CmisUpdateConflictException(CmisBaseException):
    """ UpdateConflictException """

    pass


class CmisVersioningException(CmisBaseException):
    """ VersioningException """

    pass


class NoZaakBaseFolderException(Exception):
    pass


class NoOtherBaseFolderException(Exception):
    pass


class CmisRepositoryDoesNotExist(Exception):
    pass
